#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Main GUI entrance.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import sys
import logging
from PyQt5 import QtWidgets
from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import Qt, QTimer
import _MainWindow

__version__='v0.1alpha'

LOG_CONFIG={
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                #'format': '<%(filename)s-%(funcName)s()>: %(asctime)s,%(levelname)s: %(message)s'},
                'format': '<%(filename)s-%(funcName)s()>: %(levelname)s: %(message)s'},
            },
        'handlers': {
            'default': {
                'level': 'DEBUG',
                'class': 'logging.FileHandler',
                'formatter': 'standard',
                'filename': 'MTT.log',
                },
            },
        'root' : {
                'handlers': ['default'],
                'level': 'DEBUG',
                'propagate': True
            }
        }



# TODO:
# [NO] show docs in sub folders?
# [y] fold long fields in meta data tab?
# [y] create bib texts when clicking into the bibtex tab and changing doc
# [y] add icons to folders
# doc types, books, generic etc
# [NO] insert images to note?
# [y] add add folder functionality
# [indirectly y] add add doc functionalities, by doi
# [y] add add doc functionalities, by bib
# [y] add add doc functionalities, by RIS
# import from Mendeley, zotero, Endnote?
# [y] autosave
# auto backup
# export to text (clipboard, styles), bibtex, ris. citation styles things.
# [y] collapse side tab
# [y] seperate libraries
# [y] use resource file to load icons/images
# in note tab, add time stamps at left margin?
# [y] change meta dict key naming convention to distinguish string and list types:
#   e.g. authors -> authors_l, tag -> tag_l
# possible issue with local time with added time
# [y] add logger
# [y] get all bib entries for multiple selected docs?
# [y] right click menus
# [y] option menu
# [y] import/export menu
# [y] add trash can
# [y] sqlite text search: https://stackoverflow.com/questions/35020797/how-to-use-full-text-search-in-sqlite3-database-in-django
# PDF preview
# add doc strings!!
# [partially y] make long actions threaded
# [y] need to deal with folder changes in sqlite
# [y] add doc drag drop to folders
# [y] change needs review states.
# choose pdf viewer software.
# [y] add doi lookup button
# add option to set autosave and auto backup
# add some actions to Edit menu
# [y] disable the add file button when no doc is selected, but not when adding new mannualy
# sqlite operations is restricted to a single thread.
# [y] add open doc folder action to right menu: 'xdg-mime query default inode/directory | sed 's/.desktop//g' -> e.g. nemo
# [y] auto open last datebase on launch
# rename file when exporting. Need to deal with switching to renaming after some files have been copied without renaming, or changing renaming pattern.
# perform duplicate check on adding
# [y] write a connected component function to replace networkx
# [y] text fold button has bug
# [N] add a copy to clipboard to messagebox showing failed tasks
# [y] connect the fail dialog's create folder button
# [y] Name a database file on creation and create collections folder under that name
# could probably make the thread call func using a decorator
# [y] possible to let user choose default add action?
# break down addActionTriggered()
# citation style stuff
# [y] change Ctrl-N shortcut to Add button. Maybe Ctrl-Shift-N for new lib.
# add more shortcuts
# qcompleter in some meta fields e.g. author, keyworgs
# [y] add search bar, result treewidget, clear frame, add folder from selection
# [Not now] add tooltip to search bar
# [y] the folder highlighting is still buggy
# add search result caching
# [y] save to sqlite before calling search
# [y] search desend into sub folders
# [y] crop file path within 255?
# [y] show number of docs and number of selected rows in status bar
# [y] import from Mendeley
# [y] show loaded lib name at title
# [y] deny renaming sys folders
# [y] make file paths relative
# [y] rel v.s. abs file path choices in bib, ris exports
# [y] del prints
# add doc table sort idx to settings


def main(args=None):

    if args is None:
        args=sys.argv[1:]

    logging.config.dictConfig(LOG_CONFIG)
    app=QtWidgets.QApplication(sys.argv)

    splash_pic=QPixmap(':/logo.png')
    splash=QtWidgets.QSplashScreen(splash_pic, Qt.WindowStaysOnTopHint)
    splash.show()
    splash.showMessage(__version__)

    QTimer.singleShot(1984, splash.close)

    app.processEvents()
    '''
    svgrenderer=QtSvg.QSvgRenderer('./recaman/recaman_box_n_33_fillline.svg.svg')
    qimage=QImage(522,305, QImage.Format_ARGB32)
    painter=QPainter(qimage)
    svgrenderer.render(painter)

    splash_pic=QPixmap.fromImage(qimage)
    print('# <__init__>: splash_pic', splash_pic)
    splash=QtWidgets.QSplashScreen(splash_pic, Qt.WindowStaysOnTopHint)
    splash.show()
    #splash.showMessage('Loading ...')

    QTimer.singleShot(3000, splash.close)

    app.processEvents()
    '''

    mainwindow=_MainWindow.MainWindow()
    sys.exit(app.exec_())

    return 0

if __name__=='__main__':

    main()





'''
This part contains functions handling actions triggered from the tool bar
buttons or search bar, including adding new doc, new folder, duplicate checking
and searching.


MeiTing Trunk

An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5 import QtWidgets
from lib import sqlitedb
from lib import bibparse, risparse
from lib import retrievepdfmeta
from lib.widgets import FailDialog, ThreadRunDialog
import logging


def _addPDF(jobid, abpath):
    """Retrieve meta data from PDF file

    Args:
        jobid (int): job id.
        abpath (str): abspath to a PDF file

    Returns:
        rec (int): 0 for success, 1 otherwise.
        jobid (int): the input jobid returned as it is.
        pdfmetaii (DocMeta): dict storing meta data retrieved from PDF.
    """

    try:
        pdfmetaii=retrievepdfmeta.getPDFMeta_pypdf2(abpath)
        pdfmetaii=retrievepdfmeta.prepareMeta(pdfmetaii)
        pdfmetaii['files_l']=[abpath,]
        rec=0
    except:
        pdfmetaii={}
        rec=1
    return rec,jobid,pdfmetaii


def checkFolderName(foldername, folderid, folder_dict):
    """Check for folder name conflicts

    Args:
        foldername (str): proposed name for folder.
        folderid (str): id of folder.
        folder_dict (dict): dict of folder info. key = folder_id,
                            value = (foldername, parent_id)

    Returns:
        0 if proposed name <foldername> doesn't confict with other folders
        with the same parent. 1 otherwise.
    """

    logger=logging.getLogger(__name__)

    toplevelids=[kk for kk,vv in folder_dict.items() if vv[1]=='-1']

    logger.debug('toplevelids = %s. folderid = %s' \
            %(toplevelids, folderid))

    if folderid in toplevelids:
        siblings=[folder_dict[ii][0] for ii in toplevelids if ii!=folderid]
    else:
        parentid=folder_dict[folderid][1]
        siblings=[]
        for kk,vv in folder_dict.items():
            if kk!=folderid and vv[1]==parentid:
                siblings.append(vv[0])

    if foldername in siblings:
        logger.info('foldername in siblings. foldername = %s. siblings = %s'\
                %(foldername, siblings))
        return 1

    return 0



class MainFrameToolBarSlots:

    #######################################################################
    #                        Tool bar button slots                        #
    #######################################################################


    @pyqtSlot(QtWidgets.QAction)
    def addActionTriggered(self, action):
        """Add new document to current folder

        Args:
            action (QAction): QAction sent by menu button.

        This is a slot to the triggered signal of the Add button in the tool
        bar, and is responsible for handling adding new document to the current
        folder via PDF, bibtex, RIS or manual inputs.
        """

        # action.data() is used in labelling the default add action.
        # if it's not None, return
        if action.data() is not None:
            self.logger.debug('action.data() = %s is not None. Return.' %action.data())
            return

        action_text=action.text()

        self.logger.info('action.text() = %s. action.data() = %s'\
                %(action_text, action.data()))

        #---------------------Add PDF---------------------
        if action_text=='Add PDF File':

            fname = QtWidgets.QFileDialog.getOpenFileNames(self,
                    'Choose a PDF file', '',
                    "PDF files (*.pdf);; All files (*)")[0]

            self.logger.info('Chosen PDF file = %s' %fname)

            if fname:
                self.doc_table.clearSelection()
                self.doc_table.setSelectionMode(
                        QtWidgets.QAbstractItemView.MultiSelection)

                joblist=list(zip(range(len(fname)), fname))

                t_dialog=ThreadRunDialog(_addPDF, joblist,
                        show_message='Adding PDF Files...',
                        max_threads=1,
                        get_results=True,
                        close_on_finish=True,
                        progressbar_style='classic',
                        parent=self)

                t_dialog.exec_()

                # collect failures
                faillist=[]
                for recii,jobidii,meta_dictii in t_dialog.results:

                    self.logger.debug('rec of t_dialog = %s. jobid = %s. meta_dict = %s'\
                            %(recii, jobidii, meta_dictii))

                    if recii==0:
                        self.updateTableData(None,meta_dictii)
                    else:
                        faillist.append(jobidii)

                fail_files=[jii[1] for jii in joblist if jii[0] in faillist]

                if len(fail_files)>0:

                    fail_docids=[]
                    for fii in fail_files:
                        metaii=sqlitedb.DocMeta()
                        metaii['files_l']=[fii,]
                        docidii=self.updateTableData(None,metaii)
                        fail_docids.append(docidii)

                    msg=FailDialog()
                    msg.setText('Oopsie.')
                    msg.setInformativeText('Failed to retrieve metadata from some files.')
                    msg.setDetailedText('\n'.join(fail_files))
                    msg.create_fail_summary.connect(lambda: self.createFailFolder(
                        'Add PDFs', fail_docids))
                    msg.exec_()

                self.logger.warning('failed list for PDF import = %s' %faillist)

                self.doc_table.setSelectionMode(
                        QtWidgets.QAbstractItemView.ExtendedSelection)


        #--------------------Add bibtex--------------------
        elif action_text=='Add Bibtex File':
            fname = QtWidgets.QFileDialog.getOpenFileName(self,
                    'Choose a bibtex file',
                    '',
                    "Bibtex files (*.bib);; All files (*)")[0]
            self.logger.info('Chosen bib file = %s' %fname)

            if fname:
                try:
                    bib_entries=bibparse.readBibFile(fname)
                    self.doc_table.clearSelection()
                    self.doc_table.setSelectionMode(
                            QtWidgets.QAbstractItemView.MultiSelection)
                    for eii in bib_entries:
                        self.updateTableData(None,eii)
                except Exception as e:
                    self.logger.exception('Failed to parse bib file.')

                    msg=QtWidgets.QMessageBox()
                    msg.setIcon(QtWidgets.QMessageBox.Critical)
                    msg.setWindowTitle('Error')
                    msg.setText('Oopsie. Failed to parse bib file.')
                    msg.setInformativeText('bibtexparser complaints:\n\n%s' %str(e))
                    msg.exec_()
                finally:
                    self.doc_table.setSelectionMode(
                            QtWidgets.QAbstractItemView.ExtendedSelection)


        #---------------------Add RIS---------------------
        elif action_text=='Add RIS File':

            fname = QtWidgets.QFileDialog.getOpenFileName(self,
                    'Choose an RIS file',
                    '',
                    "RIS files (*.ris);; All files (*)")[0]
            self.logger.info('Chosen ris file = %s' %fname)

            if fname:
                try:
                    ris_entries=risparse.readRISFile(fname)
                    self.doc_table.clearSelection()
                    self.doc_table.setSelectionMode(
                            QtWidgets.QAbstractItemView.MultiSelection)
                    for eii in ris_entries:
                        self.updateTableData(None,eii)
                except Exception as e:
                    self.logger.exception('Failed to parse RIS file.')

                    msg=QtWidgets.QMessageBox()
                    msg.setIcon(QtWidgets.QMessageBox.Critical)
                    msg.setWindowTitle('Error')
                    msg.setText('Oopsie. Failed to parse RIS file.')
                    msg.setInformativeText('RISparser complaints:\n\n%s' %str(e))
                    msg.exec_()
                finally:
                    self.doc_table.setSelectionMode(
                            QtWidgets.QAbstractItemView.ExtendedSelection)


        #-------------------Add manually-------------------
        elif action_text=='Add Entry Manually':

            dummy=sqlitedb.DocMeta()
            self.updateTableData(None, dummy)

        return


    @pyqtSlot(QtWidgets.QAction)
    def addFolderButtonClicked(self, action):
        """Add new folder

        Args:
            action (QAction): QAction sent by menu button.

        This is a slot to the triggered signal of the Add Folder button in the
        tool bar, and is responsible for handling adding new folder to the
        folder tree.
        """

        item=self._current_folder_item
        if item:
            folderid=item.data(1,0)

            # create new item
            if len(self.folder_dict.keys())==0:
                newid='0'
            else:
                current_ids=map(int,self.folder_dict.keys())
                newid=str(max(current_ids)+1)
            newitem=QtWidgets.QTreeWidgetItem(['New folder',str(newid)])
            style=QtWidgets.QApplication.style()
            diropen_icon=style.standardIcon(QtWidgets.QStyle.SP_DirOpenIcon)
            newitem.setIcon(0,diropen_icon)
            newitem.setFlags(newitem.flags() | Qt.ItemIsEditable)

            action_text=action.text().replace('&','') # remove shortcut symbol

            if action_text=='Create Folder':
                toplevelids=[self.libtree.topLevelItem(jj).data(1,0) for jj\
                        in range(self.libtree.topLevelItemCount())]

                if folderid in toplevelids:
                    self.libtree.addTopLevelItem(newitem)
                    parentid='-1'
                    self.logger.info('action.text() = %s. As top level folder' %action.text())
                else:
                    item.parent().addChild(newitem)
                    parentid=item.parent().data(1,0)
                    self.logger.info('action.text() = %s. As subfolder' %action.text())

            elif action_text=='Create Sub Folder':
                item.addChild(newitem)
                parentid=folderid
                self.logger.info('action.text() = %s. As subfolder' %action.text())

            self.libtree.scrollToItem(newitem)
            self.folder_dict[newid]=('New folder',parentid)
            if newid not in self.folder_data:
                self.folder_data[newid]=[]

            # let user edit folder name
            self.libtree.editItem(newitem)

            self.logger.info('New folder id = %s' %newid)

        return


    @pyqtSlot(QtWidgets.QTreeWidgetItem, int)
    def addNewFolderToDict(self, item, column):
        """Add newly create folder to folder dict

        Args:
            item (QTreeWidgetItem): newly created item for folder.
            column (int): column of change. Not used.

        This is a slot to the libtree.itemChanged signal, which is emitted when
        data associated with a QTreeWidgetItem is changed. NOTE that by default
        this covers ANY data change in an item, including its text, icon,
        background color etc.. I need to restrict this signal firing to only
        text changes (renaming folders). This is achieved by connecting
        this signal in the commitData() method of the QTreeWidget (see
        lib/widgets/folder_tree.py), and after emitting, disconnect immediately.
        Therefore this is called whenever libtree.editItem() is called, for
        instance in addFolderButtonClicked(), to update the changed folder
        name to the folder_dict.
        """

        foldername,folderid=item.data(0,0), item.data(1,0)
        self.logger.info('foldername = %s. folder id = %s'\
                %(foldername, folderid))

        if folderid not in ['-1', '-2', '-3']:
            fnameold,parentid=self.folder_dict[folderid]
            self.logger.debug('Old folder name = %s. parentid = %s'\
                    %(fnameold, parentid))

            # add new folder
            if folderid not in self.folder_data:
                self.folder_data[folderid]=[]

            # check validity of new name
            valid=checkFolderName(foldername,folderid,self.folder_dict)
            if valid!=0:

                self.logger.warning('Found invalid folder name: %s' %foldername)
                msg=QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Critical)
                msg.setWindowTitle('Name conflict')
                msg.setText('Name conflict.')
                msg.setInformativeText('The given name\n\t%s\nconflicts with another folder name.\nPlease choose another name.' %foldername)
                msg.exec_()

                item.setData(0,0,fnameold)
                # let user input again
                self.libtree.editItem(item)

                return

            # if name valid, update to dict
            self.folder_dict[folderid]=[foldername,parentid]
            self.logger.info('Added new folder name = %s. parentid = %s'\
                    %(self.folder_dict[folderid][0], self.folder_dict[folderid][1]))

        self.sortFolders()
        self.libtree.setCurrentItem(item)
        self.changed_folder_ids.append(folderid)

        return


    def sortFolders(self):
        '''Sort folders in alphabetic order while keeping sys folder at top'''

        def moveItemToTop(item):
            idx=self.libtree.indexOfTopLevelItem(item)
            item=self.libtree.takeTopLevelItem(idx)
            self.libtree.insertTopLevelItem(0,item)

        self.libtree.sortItems(0,Qt.AscendingOrder)

        # move system folders to back to top
        for itemii in reversed(self.sys_folders):
            moveItemToTop(itemii)

        return


    @pyqtSlot()
    def checkDuplicateClicked(self):
        """Check duplicates within current folder

        This is a slot to the triggered signal of the Check Duplicate button in
        the tool bar.
        """

        docids=self._current_docids
        self.doc_table.setVisible(False)

        current_folder=self._current_folder[0]
        self.duplicate_result_frame.clear_duplicate_label.setText(
                'Checking duplicates in folder "%s".' %current_folder)

        self.duplicate_result_frame.checkDuplicates(self.meta_dict,
                self._current_folder,
                docids,
                None)

        return


    @pyqtSlot(QtWidgets.QTreeWidgetItem,QtWidgets.QTreeWidgetItem)
    def duplicateResultCurrentChange(self, current, previous):
        """Select a doc in the result of duplicate checking

        Args:
            current (QTreeWidgetItem): item for a doc in the result tree widget.
            previous (QTreeWidgetItem): not used.

        """

        if current:
            docid=int(current.data(6,0))
            self.logger.info('current doc id = %s' %docid)
            self.loadMetaTab(docid)
            self.loadBibTab(docid)
            self.loadNoteTab(docid)

        return


    @pyqtSlot()
    def searchBarClicked(self):
        """Start search in the database

        This is a slot to the clicked signal of the search button in the tool
        bar, and to the returnPressed signal of the search_bar.

        Search text is retrieved from the search bar, and search options
        are collected from the states of the search button menu. Then
        these are passed as input args to the search function defined in
        lib/widgets/search_res_frame.py

        NOTE that as search is done in the sqlite data, a saving is first
        called.
        """

        text=self.search_bar.text()
        self.logger.info('Searched term = %s' %text)

        if len(text)==0:
            return

        #-----------Get search fields and option-----------
        menu=self.search_button.menu()
        actions=menu.findChildren(QtWidgets.QWidgetAction)
        new_search_fields=[]

        for actii in actions:
            wii=actii.defaultWidget()
            self.logger.info('action %s isChecked() = %s'\
                    %(actii.text(), wii.isChecked()))

            if actii.text()=='Include sub-folders':
                desend=wii.isChecked()
            else:
                if wii.isChecked():
                    new_search_fields.append(actii.text())

        self.settings.setValue('search/search_fields',new_search_fields)
        self.settings.setValue('search/desend_folder',desend)

        if len(new_search_fields)==0:
            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setWindowTitle('Information')
            msg.setText('Input needed')
            msg.setInformativeText('Select at least one field to search')
            msg.exec_()
            return

        current_folder=self._current_folder

        # NOTE: need to write to sqlite before searching
        self.saveToDatabase()

        # NOTE: order matters here:
        self.status_bar.showMessage('Searching ...')
        self.doc_table.setVisible(False)
        self.search_res_frame.search(self.db, text, new_search_fields,
                current_folder[1], self.meta_dict, desend)

        return


    @pyqtSlot(str,list)
    def createFolderFromSearch(self, search_text, docids):
        """Create a folder and populate with docs selected from search results

        Args:
            search_text (str): searched text in the search bar.
            docids (list): ids of docs to add to the new folder


        This is a slot to the create_folder_sig emitted by the
        search_res_frame. It will create a top level folder with the name
        'search_<search_text>', and add selected docs to it for later use.
        """

        foldername='search_%s' %search_text
        toplevelfolders=[vv[0] for kk,vv in self.folder_dict.items() if\
                vv[1] in ['-1',]]

        self.logger.info('foldername = %s' %foldername)
        self.logger.debug('toplevel folders = %s' %toplevelfolders)

        # rename till no conflict
        append=1
        while foldername in toplevelfolders:
            foldername='%s_%d' %(foldername,append)
            append+=1
            self.logger.debug('new foldername after renaming = %s' %foldername)

        # create new item
        current_ids=map(int,self.folder_dict.keys())
        newid=str(max(current_ids)+1)
        newitem=QtWidgets.QTreeWidgetItem([foldername,str(newid)])
        style=QtWidgets.QApplication.style()
        diropen_icon=style.standardIcon(QtWidgets.QStyle.SP_DirOpenIcon)
        newitem.setIcon(0,diropen_icon)
        newitem.setFlags(newitem.flags() | Qt.ItemIsEditable)

        self.folder_dict[newid]=(foldername,'-1')
        self.folder_data[newid]=docids
        self.libtree.addTopLevelItem(newitem)
        self.libtree.scrollToItem(newitem)
        self.libtree.setCurrentItem(newitem)

        self.logger.info('New folder id=%s' %newid)
        self.changed_folder_ids.append(newid)

        return


    @pyqtSlot()
    def hideDocTable(self):
        '''Not in use'''
        if self.doc_table.isVisible():
            self.doc_table.setVisible(False)
        return


    @pyqtSlot(QtWidgets.QTreeWidgetItem,QtWidgets.QTreeWidgetItem)
    def searchResultCurrentChange(self,current,previous):
        """Select a doc in the search results

        Args:
            current (QTreeWidgetItem): item for a doc in the result tree widget.
            previous (QTreeWidgetItem): not used.

        """

        if current:
            docid=int(current.data(5,0))
            self.logger.info('current doc id = %s' %docid)
            self.loadMetaTab(docid)
            self.loadBibTab(docid)
            self.loadNoteTab(docid)

        return









'''
Export PDFs with annotations. This is used in importing Mendeley data.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import PyPDF2
from . import pdfannotation
import logging

LOGGER=logging.getLogger(__name__)


def exportPdf(fin, abpath_out, annotations):
    '''Export PDF with annotations.

    Args:
        fin (str): abspath to input PDF file.
        abpath_out (str): abspath to output PDF file.
        annotations (dict): annotation info. See import_mendeley.py
                            getHighlights() for more info.
    '''

    try:
        inpdf = PyPDF2.PdfFileReader(open(fin, 'rb'))
        if inpdf.isEncrypted:
            # PyPDF2 seems to think some files are encrypted even
            # if they are not. We just ignore the encryption.
            # This seems to work for the one file where I saw this issue
            #inpdf._override_encryption = True
            #inpdf._flatten()
            # UPDATE: trying to decrypt takes a lot of time,
            # as this rarely happens to academic docs I'm skipping this
            # and simply treat as fail
            #raise Exception("Skip encrypt")
            return
    except IOError:
        LOGGER.warning('Could not open pdf file %s' %fin)

    # retain meta data
    meta = inpdf.getDocumentInfo()
    outpdf = PyPDF2.PdfFileWriter()
    outpdf.addMetadata(meta)

    highlights=annotations.get('highlights',None)
    if highlights is None:
        hlpages=[]
    else:
        hlpages=list(highlights.keys())
        hlpages.sort()

    notes=annotations.get('notes',None)
    if notes is None:
        ntpages=[]
    else:
        ntpages=list(notes.keys())
        ntpages.sort()

    #----------------Loop through pages----------------
    pages=range(1,inpdf.getNumPages()+1)

    for pii in pages:
        inpg = inpdf.getPage(pii-1)

        #----------------Process highlights----------------
        if pii in hlpages:
            for hjj in highlights[pii]:
                # Changes suggested by matteosecli: add author of highlight:
                anno = pdfannotation.createHighlight(hjj["rect"],
                        author=hjj['author'],
                        cdate=hjj["cdate"], color=hjj['color'])
                inpg=pdfannotation.addAnnotation(inpg,outpdf,anno)

        #------------------Process notes------------------
        if pii in ntpages:
            for njj in notes[pii]:
                note = pdfannotation.createNote(njj["rect"], \
                        contents=njj["content"], author=njj["author"],\
                        cdate=njj["cdate"])
                inpg=pdfannotation.addAnnotation(inpg,outpdf,note)

        outpdf.addPage(inpg)

    #-----------------------Save-----------------------
    if os.path.isfile(abpath_out):
        os.remove(abpath_out)

    with open(abpath_out, mode='wb') as fout:
        outpdf.write(fout)

    LOGGER.debug('Exported annotated pdf.')

    return




'''
About dialog.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from PyQt5.QtGui import QPixmap
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QDialogButtonBox
from PyQt5.QtCore import Qt
import logging

LOGGER=logging.getLogger(__name__)
REPO_URL='https://github.com/Xunius/MeiTingTrunk'



class AboutDialog(QtWidgets.QDialog):

    def __init__(self, version, parent=None):

        super(self.__class__,self).__init__(parent=parent)

        self.version=version

        logo=QPixmap(':/logo.png')
        logo_label=QtWidgets.QLabel(self)
        logo_label.setPixmap(logo)

        self.setWindowTitle('About MeiTing Trunk')

        va=QtWidgets.QVBoxLayout(self)
        va.addWidget(logo_label)

        label=QtWidgets.QLabel('MeiTing Truck')
        label.setStyleSheet('font: bold')
        va.addWidget(label, 0 , Qt.AlignHCenter)
        va.addWidget(QtWidgets.QLabel(self.version), 0 , Qt.AlignHCenter)
        va.addWidget(QtWidgets.QLabel('An open source reference management tool developed in PyQt5 and Python3.'), 0 , Qt.AlignHCenter)
        va.addWidget(QtWidgets.QLabel('Copyright 2018-2019 Guang-zhi XU'), 0 , Qt.AlignHCenter)
        va.addWidget(QtWidgets.QLabel('Please consider contributing. Repo:'),
                0 , Qt.AlignHCenter)
        va.addWidget(QtWidgets.QLabel(REPO_URL), 0 , Qt.AlignHCenter)

        self.buttons=QDialogButtonBox(QDialogButtonBox.Ok, Qt.Horizontal, self)
        self.buttons.accepted.connect(self.accept)

        va.addWidget(self.buttons)










'''
This part contains methods dealing with actions on documents in the doc table,
including selection, right clicking menu, deletion, opening.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import subprocess
from PyQt5.QtCore import Qt, pyqtSignal, QTimer, pyqtSlot, QModelIndex
from PyQt5 import QtWidgets
from PyQt5.QtGui import QFont, QBrush, QColor, QCursor, QIcon
from lib import sqlitedb
from lib import bibparse, risparse
from lib.tools import parseAuthors
from lib.widgets import Master, FailDialog


class MainFrameDocTableSlots:

    #######################################################################
    #                           Doc table slots                           #
    #######################################################################

    @pyqtSlot()
    def docTableClicked(self):

        self.logger.debug('Doc table clicked. Set to extendedselection.')
        self.doc_table.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        # count selected docs
        sel_rows=self.doc_table.selectionModel().selectedRows()
        if len(sel_rows)>1:
            self.status_bar.showMessage('%d rows selected. %d rows in total'\
                    %(len(sel_rows), self.doc_table.model().rowCount(None)))

        return


    @pyqtSlot(QModelIndex, QModelIndex)
    def modelDataChanged(self, index1, index2):
        """Update favourite and read dict data in response to checkbox changes

        Args:
            index1 (QModelIndex): index of changed row
            index2 (QModelIndex): index of changed row?

        This is a slot to doc_table.model().dataChanged signal, emitted when
        the checkboxes for favourite and read columns change state (in response
        to user clicking).
        """

        assert index1==index2, 'only one doc row is changed'
        # NOTE that the row is in general different from
        # self.doc_table.currentIndex().row(), the former may not change on
        # clicking the checkboxes
        row=index1.row()

        docid=self._tabledata[row][0]
        fav=self._tabledata[row][1].isChecked()
        read=self._tabledata[row][2].isChecked()

        self.meta_dict[docid]['favourite']='true' if fav else 'false'
        self.meta_dict[docid]['read']='true' if read else 'false'

        self.logger.info('Changed row = %s. Changed docid = %s. meta_dict["favourite"] = %s. meta_dict["read"] = %s' \
                %(row, docid, self.meta_dict[docid]['favourite'],\
                self.meta_dict[docid]['read']))

        self.changed_doc_ids.append(docid)

        return


    @pyqtSlot(QModelIndex, QModelIndex)
    def selDoc(self, current, previous):
        """Select a doc in the doc table

        Args:
            current (QModelIndex): index of current selection
            previous (QModelIndex): index of previous selection

        This is a slot to the doc_table.selectionModel().currentChanged signal.
        """

        rowid=current.row()
        docid=self._tabledata[rowid][0]
        self.logger.info('Selected rowid = %s. docid = %s' %(rowid, docid))

        self.loadMetaTab(docid)
        self.loadBibTab(docid)
        self.loadNoteTab(docid)

        #------------Remove highlights for all folders-------
        self.removeFolderHighlights()

        #-------------------Get folders-------------------
        folders=self.meta_dict[docid]['folders_l']
        folders=[str(fii[0]) for fii in folders]
        self.logger.debug('Ids of folders containing doc (%s): %s' %(docid, folders))

        #---------Highlight folders contaning doc---------
        hi_color=self.settings.value('display/folder/highlight_color_br',
                QBrush)
        for fii in folders:
            mii=self.libtree.findItems(fii, Qt.MatchExactly | Qt.MatchRecursive,
                    column=1)
            if len(mii)>0:
                for mjj in mii:
                    mjj.setBackground(0, hi_color)

        #------------Show confirm review frame------------
        if self.meta_dict[docid]['confirmed'] in [None, 'false']:
            self.confirm_review_frame.setVisible(True)
        else:
            self.confirm_review_frame.setVisible(False)

        return


    def docTableMenu(self, pos):
        """Right click menu in the doc table

        Deletion actions differ depending on which folder the current doc is in:

            * in normal folder: "delete from folder" and "delete from library"
                  actions are added.
            * in Needs Review folder: no deletion action added. Docs can only
                  be removed from this folder by confirming
                  their meta data.
            * in All folder: "delete from library" action is added.
            * in Trash folder, or any folder within Trash: "delete from Trash"
                  action is added.

        "Mark as Needs Review" is not added when in Needs Review folder.
        "Open File Externally" and "Open Containing Folder" are disabled when
        none selected docs have any associated file.

        """

        menu=QtWidgets.QMenu()
        current_folderid=self._current_folder_item.data(1,0)
        trashed_folder_ids=self._trashed_folder_ids+['-3']

        open_action=menu.addAction('&Open File Externally')
        open_action.setIcon(QIcon.fromTheme('document-open'))
        open_action.setShortcut('O')

        open_folder_action=menu.addAction('Open Containing &Folder')
        #open_folder_action.setIcon(QIcon.fromTheme('system-file-manager'))
        open_folder_action.setIcon(self.style().standardIcon(
            QtWidgets.QStyle.SP_DirIcon))
        open_folder_action.setShortcut('F')

        #-----------------Deletion actions-----------------
        del_from_folder_action=QtWidgets.QAction('&Delete From Current Folder',
                menu)
        del_from_folder_action.setIcon(QIcon.fromTheme('user-trash'))
        del_from_folder_action.setShortcut('D')

        del_from_lib_action=QtWidgets.QAction('Delete From Library',menu)
        del_from_lib_action.setIcon(QIcon.fromTheme('user-trash'))

        del_from_trash_action=QtWidgets.QAction('Delete From Trash',menu)
        del_from_trash_action.setIcon(QIcon.fromTheme('edit-delete'))

        if current_folderid=='-1':
            menu.addAction(del_from_lib_action)
        elif current_folderid=='-2':
            pass
        elif current_folderid in trashed_folder_ids:
            menu.addAction(del_from_trash_action)
        else:
            menu.addAction(del_from_folder_action)
            menu.addAction(del_from_lib_action)

        #-------------Mark needs review action-------------
        mark_needsreview_action=QtWidgets.QAction('&Mark document as Needs Review',menu)
        mark_needsreview_action.setIcon(self.style().standardIcon(
            QtWidgets.QStyle.SP_MessageBoxInformation))
        mark_needsreview_action.setShortcut('M')

        if current_folderid!='-2':
            menu.addAction(mark_needsreview_action)

        #--------------Duplicate check action--------------
        check_duplicate_folder_action=menu.addAction('Check Du&plicates Within Folder')
        check_duplicate_folder_action.setIcon(QIcon.fromTheme('edit-find'))
        check_duplicate_folder_action.setShortcut('P')

        check_duplicate_lib_action=menu.addAction('Check Duplicates Within Library')
        check_duplicate_lib_action.setIcon(QIcon.fromTheme('edit-find'))

        #------------------Export actions------------------
        menu.addSeparator()
        export_bib_action=menu.addAction('Export to &bibtex File')
        export_bib_action.setIcon(QIcon.fromTheme('document-save-as'))
        export_bib_action.setShortcut('B')

        export_ris_action=menu.addAction('Export to &RIS File')
        export_ris_action.setIcon(QIcon.fromTheme('document-save-as'))
        export_ris_action.setShortcut('R')

        copy_clipboard_action=menu.addAction('Export Citation To &Clipboard')
        copy_clipboard_action.setIcon(QIcon.fromTheme('edit-copy'))
        copy_clipboard_action.setShortcut('C')

        sel_rows=self.doc_table.selectionModel().selectedRows()
        sel_rows=[ii.row() for ii in sel_rows]

        if len(sel_rows)>0:

            docids=[self._tabledata[ii][0] for ii in sel_rows]
            has_files=[self.meta_dict[docii]['has_file'] for docii in docids]

            self.logger.info('Selected docids = %s. has_files = %s'\
                    %(docids, has_files))

            if any(has_files):
                open_action.setEnabled(True)
                open_folder_action.setEnabled(True)
            else:
                open_action.setDisabled(True)
                open_folder_action.setDisabled(True)

            action=menu.exec_(QCursor.pos())

            if action:
                self.logger.info('action.text() = %s' %action.text())

                if action==open_action:
                    open_docs=[docids[ii] for ii in range(len(docids)) if has_files[ii]]
                    self.openDoc(open_docs)

                elif action==open_folder_action:
                    open_docs=[docids[ii] for ii in range(len(docids)) if has_files[ii]]
                    self.openDocFolder(open_docs)

                elif action==del_from_folder_action:
                    foldername,folderid=self._current_folder
                    self.delFromFolder(docids, foldername, folderid, True)

                elif action==del_from_lib_action:
                    self.delDoc(docids,True)

                elif action==del_from_trash_action:
                    self.destroyDoc(docids,current_folderid,True,True)

                elif action==mark_needsreview_action:
                    self.markDocNeedsReview(docids)

                elif action==check_duplicate_folder_action:
                    self.checkDocDuplicate(docids,'folder')

                elif action==check_duplicate_lib_action:
                    self.checkDocDuplicate(docids,'lib')

                elif action==export_bib_action:
                    self.exportToBib(docids)

                elif action==export_ris_action:
                    self.exportToRIS(docids)

                elif action==copy_clipboard_action:
                    self.copyToClipboard(docids,style=None)

        return


    def openDoc(self, docids):
        """Open attached files externally

        Args:
            docids (list): ids of docs to open

        xdg-open is called to open the attached files. If return code of this
        subprocess is 0, set 'read' dict value to 'true'.

        NOTE: need to re-write this for Windows version.
        """

        self.logger.info('docids = %s' %docids)
        lib_folder=self.settings.value('saving/current_lib_folder',str)

        for docii in docids:
            file_pathii=self.meta_dict[docii]['files_l'][0] # take the 1st file
            # prepend relative path with folder path
            file_pathii=os.path.join(lib_folder,file_pathii)

            if not os.path.exists(file_pathii):
                msg=QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Warning)
                msg.setWindowTitle('Error')
                msg.setText("Can't find file.")
                msg.setInformativeText("No such file: %s. Please re-attach the document file." %file_pathii)
                msg.exec_()
                #return

            self.logger.debug('docid = %s. file_path = %s' %(docii, file_pathii))

            prop=subprocess.call(('xdg-open', file_pathii))
            if prop==0:
                # set read to True
                self.meta_dict[docii]['read']='true'
                self.changed_doc_ids.append(docii)

        # refresh to show read change
        self.loadDocTable(folder=self._current_folder,sortidx=4,
                sel_row=None)

        return


    def openDocFolder(self, docids):
        """Locate the attached files in the file manager

        Args:
            docis (list): ids of docs to locate files

        xdg-mime is used to get the system file manager, which is then launched
        in subprocess.

        NOTE: need to re-write this for Windows verion.
        """

        self.logger.info('docids = %s' %docids)

        #------------Get default file mananger------------
        prop=subprocess.Popen(['xdg-mime','query','default','inode/directory'],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        file_man=prop.communicate()[0].decode('ascii').strip().replace('.desktop','')

        #----------------Open file manager----------------
        lib_folder=self.settings.value('saving/current_lib_folder',str)
        for docii in docids:
            file_pathii=self.meta_dict[docii]['files_l'][0] # take the 1st file
            file_pathii=os.path.join(lib_folder,file_pathii)

            self.logger.debug('docid = %s. file_path = %s' %(docii, file_pathii))

            prop=subprocess.call((file_man, file_pathii))

        return


    @pyqtSlot(list, str, str, bool)
    def delFromFolder(self, docids, foldername, folderid, reload_table):
        """Delete a doc from a folder

        Args:
            docids (list): ids of docs to delete.
            foldername (str): name of folder to delete docs from.
            folderid (str): id of folder to delete docs from.
            reload_table (bool): whether to reload the doc table after deletion.

        This is the implementation of the "Delete From Folder" action in the
        right click menu of doc table. See docTableMenu().

        It is also a slot to the del_doc_from_folder_signal emitted when the
        result frame of duplicate checking is shown and user chooses to delete
        some duplicated docs from the folder.

        Deleted docs become orphan docs if they no longer appear in any other
        normal folder, and will be put to the Trash folder.
        """

        self.logger.info('docids = %s. foldername = %s. folderid = %s'\
                %(docids, foldername, folderid))

        # remove doc from folder
        for idii in docids:
            self.folder_data[folderid].remove(idii)
            self.logger.debug('folder_data[folderid] = %s' %self.folder_data[folderid])

            # remove folder from doc
            if (int(folderid),foldername) in self.meta_dict[idii]['folders_l']:
                self.meta_dict[idii]['folders_l'].remove((int(folderid),foldername))
                self.changed_doc_ids.append(idii)
                self.logger.debug("meta_dict['folders_l'] = %s"\
                        %self.meta_dict[idii]['folders_l'])

        # check orphan
        orphan_docs=sqlitedb.findOrphanDocs(self.folder_data,docids,
                self._trashed_folder_ids)

        for idii in orphan_docs:
            self.meta_dict[idii]['deletionPending']='true'
            self.folder_data['-3'].append(idii)
            self.changed_doc_ids.append(idii)

            self.logger.debug('Set orphan doc deletionPending to: %s'\
                    %self.meta_dict[idii]['deletionPending'])

        if reload_table:
            self.loadDocTable(folder=(foldername,folderid),sel_row=None)


        return


    @pyqtSlot(list, bool)
    def delDoc(self, docids, reload_table):
        """Delete docs from the library (across all folders)

        Args:
            docids (list): ids of docs to delete.
            reload_table (bool): whether to reload the doc table after deletion.

        This is the implementation of the "Delete From Library" action in the
        right click menu of the doc table. See docTableMenu().

        It is also a slot to the del_doc_from_lib_signal emitted when the
        result frame of duplicate checking is shown and user chooses to delete
        some duplicated docs from all folders.

        Docs are removed from all normal folders, and thus become orphan docs
        and are put to the Trash folder.
        """

        self.logger.info('docids = %s' %docids)

        choice=QtWidgets.QMessageBox.question(self, 'Confirm deletion',
                'Confirm deleting a document from library?',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        if choice==QtWidgets.QMessageBox.Yes:

            for idii in docids:

                # remove from all folders
                for kk,vv in self.folder_data.items():
                    if idii in vv:
                        vv.remove(idii)

                        self.logger.debug('doc %s in folder_data[%s]?: %s'\
                                %(idii, kk, idii in self.folder_data[kk]))

                self.meta_dict[idii]['folders_l']=[]
                self.folder_data['-3'].append(idii)
                self.meta_dict[idii]['deletionPending']='true'
                self.changed_doc_ids.append(idii)

                self.logger.debug('Set orphan doc deletionPending to: %s'\
                        %self.meta_dict[idii]['deletionPending'])

            if reload_table:
                self.loadDocTable(folder=self._current_folder,sel_row=None)

        return


    def markDocNeedsReview(self, docids):
        """Set the confirmed dict value to 'false'

        Args:
            docids (list): ids of docs to change.

        Changed docs are added to the Needs Review folder.
        """

        self.logger.info('docids = %s' %docids)

        for idii in docids:
            self.meta_dict[idii]['confirmed']='false'
            if idii not in self.folder_data['-2']:
                self.folder_data['-2'].append(idii)
            self.changed_doc_ids.append(idii)

            self.logger.debug('Set doc confirmed to: %s. Doc in folder_data[-2]: %s'\
                    %(self.meta_dict[idii]['confirmed'],
                        idii in self.folder_data['-2']))

        row=self.doc_table.currentIndex().row()
        # didn't find a better way to refresh the table view
        self.loadDocTable(folder=self._current_folder,sortidx=4,
                sel_row=row)

        return


    def destroyDoc(self, docids, current_folderid, ask, reload_table):
        """Delete docs from trash

        Args:
            docids (list): ids of docs to delete.
            current_folderid (str): id of current selected folder.
            ask (bool): whether to prompt for confirmation.
            reload_table (bool): whether to reload the doc table afterwards.

        This is the implementation of the "Delete From Trash" action in the
        right click menu of the doc table. See docTableMenu().

        It is also called when deleting folder(s) from Trash, see
        delFolderFromTrash().

        "destroy" means deleting from Trash, and after committing to sqlite,
        becomes permanently gone. Therefore it is only called in actions
        within the Trash context.

        NOTE that if a doc appears in Trash ITSELF, it is by design an
        orphan doc (no longer appearing in any other normal folder). But
        if it appears in a folder somewhere in Trash, it is not necessarily
        an orphan, it may still exists in a normal folder outside of
        Trash. In such cases, it is equivalently an "Delete From Folder" action
        in that it is only removed from the current folder. If it is indeed
        an orphan doc, it will be removed from the in-memory meta dict
        self.meta_dict, from all normal folders (self.folder_data), and
        after committing to sqlite, be permanently deleted.

        """

        self.logger.info('docids = %s' %docids)

        if ask:
            choice=QtWidgets.QMessageBox.question(self, 'Confirm deletion',
                    'Confirm deleting a document permanently?',
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        if not ask or (ask and choice==QtWidgets.QMessageBox.Yes):

            current_foldername=self.folder_dict[current_folderid][0]

            for idii in docids:

                # NOTE!!: if doc still exists in some other folder outside Trash,
                # (which can be tell by deletionPending=='false'), don't del,
                # only remove from current folder.

                if self.meta_dict[idii]['deletionPending']=='false':

                    self.logger.warning('Doc still relevant. Only delete from current folder.')

                    if idii in self.folder_data[current_folderid]:
                        self.folder_data[current_folderid].remove(idii)
                        self.meta_dict[idii]['folders_l'].remove(
                                (int(current_folderid), current_foldername))

                        self.logger.debug('doc %s in folder_data[%s]?: %s'\
                                %(idii, current_folderid,
                                    idii in self.folder_data[current_folderid]))
                        self.logger.debug("current folder id = %s. meta_dict[%s]['folders_l'] = %s" \
                                %(current_folderid, idii,
                                    self.meta_dict[idii]['folders_l']))

                    continue

                # remove from all folders
                for kk,vv in self.folder_data.items():
                    if idii in vv:
                        vv.remove(idii)

                        self.logger.debug('doc %s in folder_data[%s]?: %s'\
                                %(idii, kk, idii in self.folder_data[kk]))

                self.changed_doc_ids.append(idii)
                del self.meta_dict[idii]

                self.logger.info('Deleted %s from meta_dict' %idii)

            if reload_table:
                self.loadDocTable(folder=self._current_folder,sel_row=None)

        return



    def exportToBib(self, docids):
        """Export meta data of docs to bibtex file

        Args:
            docids (list): ids of docs to export.

        Multiple entries are saved to a same bib file.
        Failed jobs will be shown in a pop-up dialog.
        """

        self.logger.info('docids = %s' %docids)

        if len(docids)==1:
            default_path='./%s.bib' %(self.meta_dict[docids[0]]['citationkey'])
        else:
            default_path='./bibtex.bib'
        self.logger.debug('Default export path = %s' %default_path)

        fname = QtWidgets.QFileDialog.getSaveFileName(self,
                'Save Citaitons to bib File',
                default_path,
                "bib Files (*.bib);; All files (*)")[0]
        self.logger.info('Chosen bib file = %s' %fname)

        def saveBib(results):
            '''Collect results and write file'''
            faillist=[]

            text=''
            for recii,jobii,textii,docii in results:
                if recii==0:
                    text=text+textii+'\n'
                elif recii==1:
                    faillist.append(docii)

            with open(fname,'w') as fout:
                fout.write(text)

            # show failed jobs
            if len(faillist)>0:
                fail_entries=[]
                for docii in faillist:
                    metaii=self.meta_dict[docii]
                    entryii='* %s_%s_%s' %(', '.join(metaii['authors_l']),
                            metaii['year'],
                            metaii['title'])
                    fail_entries.append(entryii)

                msg=FailDialog()
                msg.setText('Oopsie')
                msg.setInformativeText('Failed to export some entires.')
                msg.setDetailedText('\n'.join(fail_entries))
                msg.create_fail_summary.connect(lambda: self.createFailFolder(
                    'bibtext export', faillist))
                msg.exec_()

            return


        if fname:
            # omit some fields
            omit_keys=self.settings.value('export/bib/omit_fields', [], str)
            if isinstance(omit_keys,str) and omit_keys=='':
                omit_keys=[]

            # abs v.s. relative file paths
            path_type=self.settings.value('export/bib/path_type',str)
            if path_type=='absolute':
                prefix=self.settings.value('saving/current_lib_folder',str)
            elif path_type=='relative':
                prefix=''

            job_list=[]
            for ii,docii in enumerate(docids):
                job_list.append((ii,self.meta_dict[docii],omit_keys,prefix))

            # run in separate thread
            self.master1=Master(bibparse.metaDictToBib, job_list,
                    1, self.progressbar,
                    'classic', self.status_bar, 'Exporting to bibtex...',
                    post_process_func=saveBib,
                    close_on_finish=False)
            self.master1.run()

        self.logger.info('Bib file exported.')

        return


    def exportToRIS(self,docids):
        """Export meta data of docs to RIS file

        Args:
            docids (list): ids of docs to export.

        Multiple entries are saved to a same RIS file.
        Failed jobs will be shown in a pop-up dialog.
        """

        self.logger.info('docids = %s' %docids)

        if len(docids)==1:
            default_path='%s.ris' %(self.meta_dict[docids[0]]['citationkey'])
        else:
            default_path='./RIS.ris'

        self.logger.debug('Default export path = %s' %default_path)

        fname = QtWidgets.QFileDialog.getSaveFileName(self,
                'Save Citaitons to RIS File',
                default_path,
                "bib Files (*.ris);; All files (*)")[0]
        self.logger.info('Chosen ris file = %s' %fname)

        def saveRIS(results):
            faillist=[]

            text=''
            for recii,jobii,textii,docii in results:
                if recii==0:
                    text=text+textii+'\n'
                elif recii==1:
                    faillist.append(docii)

            with open(fname,'w') as fout:
                fout.write(text)

            # show failed jobs
            if len(faillist)>0:
                fail_entries=[]
                for docii in faillist:
                    metaii=self.meta_dict[docii]
                    entryii='* %s_%s_%s' %(', '.join(metaii['authors_l']),
                            metaii['year'],
                            metaii['title'])
                    fail_entries.append(entryii)

                msg=FailDialog()
                msg.setText('Oopsie')
                msg.setInformativeText('Failed to export some entires.')
                msg.setDetailedText('\n'.join(fail_entries))
                msg.create_fail_summary.connect(lambda: self.createFailFolder(
                    'RIS export', faillist))
                msg.exec_()

            return

        if fname:
            path_type=self.settings.value('export/ris/path_type',str)
            if path_type=='absolute':
                prefix=self.settings.value('saving/current_lib_folder',str)
            elif path_type=='relative':
                prefix=''

            job_list=[]
            for ii,docii in enumerate(docids):
                job_list.append((ii,self.meta_dict[docii],prefix))
            self.master1=Master(risparse.metaDictToRIS, job_list,
                    1, self.progressbar,
                    'classic', self.status_bar, 'Exporting to RIS...',
                    post_process_func=saveRIS,
                    close_on_finish=False)
            self.master1.run()

        self.logger.info('RIS file created.')

        return


    def copyToClipboard(self, docids, style=None):
        """Copy citations of selected docs to system clipboard

        Args:
            docids (list): ids of docs to generate citation.

        Kwargs:
            style (str): citation to style. If None, use a default style

        NOTE: this is only a demo, needs to implememnt some proper style
        formatting. Help needed.
        """

        self.logger.info('docids = %s' %docids)

        cb=QtWidgets.QApplication.clipboard()
        meta={}
        meta_list=[]

        for idii in docids:
            docii=self.meta_dict[idii]

            self.logger.debug('docii["authors_l"]=%s' %docii['authors_l'])

            authorsii=parseAuthors(docii['authors_l'])[2]
            if len(authorsii)==0:
                authorsii='UNKNOWN'
            elif len(authorsii)==1:
                authorsii=authorsii[0]
            elif len(authorsii)>=2:
                authorsii=' and '.join(authorsii)

            meta['authors']=authorsii
            meta['year']=docii['year']
            meta['title']=docii['title']
            meta['journal']=docii['publication']
            meta['issue']=docii['issue']
            meta['pages']=docii['pages']

            self.logger.debug('meta = %s' %meta)

            meta_list.append(meta)


        text=''
        for docii in meta_list:
            if style is None:

                textii='%s, %s: %s. %s, <html><b>%s</b></html>, %s' %(docii['authors'],
                        docii['year'],
                        docii['title'],
                        docii['journal'],
                        docii['issue'],
                        docii['pages']
                        )

                text=text+textii+'\n\n'

        cb.setText(text)

        return


    @pyqtSlot(QModelIndex)
    def docDoubleClicked(self, idx):
        """Slot to double clicking in doc table

        Args:
            idx (QModelIndex): index of doc table item been clicked on.

        If the selected doc has only 1 attachment file, call openDoc() on it.
        Otherwise, pop-up a dialog listing all attachments, and let user choose
        which one(s) to open externally. After successful subprocess call,
        set 'read' dict value to 'true'.
        """

        row_idx=idx.row()
        self.logger.info('Clicked row=%s' %row_idx)

        docid=self._tabledata[row_idx][0]
        files=self.meta_dict[docid]['files_l']
        nfiles=len(files)

        if nfiles==0:
            return
        elif nfiles==1:
            self.openDoc([docid,])
        else:

            self.logger.info('Multiple files associated with doc. n = %d' %nfiles)

            dialog=QtWidgets.QDialog()
            dialog.resize(500,500)
            dialog.setWindowTitle('Open Files Externally')
            dialog.setWindowModality(Qt.ApplicationModal)
            layout=QtWidgets.QVBoxLayout()
            dialog.setLayout(layout)

            label=QtWidgets.QLabel('Select file(s) to open')
            label_font=QFont('Serif',12,QFont.Bold)
            label.setFont(label_font)
            layout.addWidget(label)

            listwidget=QtWidgets.QListWidget()
            listwidget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

            lib_folder=self.settings.value('saving/current_lib_folder',str)
            for fii in files:
                fii=os.path.join(lib_folder,fii)
                listwidget.addItem(fii)

            listwidget.setCurrentRow(0)
            layout.addWidget(listwidget)

            buttons=QtWidgets.QDialogButtonBox(
                QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
                Qt.Horizontal, dialog)

            buttons.accepted.connect(dialog.accept)
            buttons.rejected.connect(dialog.reject)
            layout.addWidget(buttons)

            rec=dialog.exec_()
            self.logger.debug('return value from dialog: %s' %rec)

            if rec:
                sel_files=listwidget.selectionModel().selectedRows()
                sel_files=[listwidget.item(ii.row()) for ii in sel_files]
                sel_files=[ii.data(0) for ii in sel_files]
                self.logger.info('Selected files=%s' %sel_files)

                if len(sel_files)>0:
                    for fii in sel_files:
                        prop=subprocess.call(('xdg-open',fii))
                        if prop==0:
                            # set read to True
                            self.meta_dict[docid]['read']='true'
                            self.logger.debug("New value of meta_dict[docid]['read'] = %s"\
                                    %self.meta_dict[docid]['read'])

                            self.changed_doc_ids.append(docid)

        return


    def checkDocDuplicate(self, docids, domain):
        """Search for duplicates with selected doc(s)

        Args:
            docids (list): ids of docs to check.
            domain (str): if 'folder', search within the current folder.
                          if 'lib', search within all folderes.


        """

        self.logger.info('docids = %s, domain = %s' %(docids, domain))

        self.doc_table.setVisible(False)

        if domain=='folder':
            current_folder=self._current_folder
            docids1=self._current_docids
        elif domain=='lib':
            current_folder=('All', '-1')
            docids1=list(self.meta_dict.keys())

        self.duplicate_result_frame.clear_duplicate_label.setText(
                'Checking duplicates in folder "%s".' %current_folder[0])

        self.duplicate_result_frame.checkDuplicates(self.meta_dict,
                current_folder, docids1, docids)

        return




'''
Contains a few functions manipulating the meta tab.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from PyQt5.QtCore import pyqtSlot


class MainFrameMetaTabSlots:

    #######################################################################
    #                            Meta tab slots                           #
    #######################################################################

    def clearMetaTab(self):

        for kk,vv in self._current_meta_dict.items():
            if kk=='files_l':
                self.t_meta.delFileField()
            else:
                vv.clear()
                vv.setReadOnly(True)

        for tii in [self.note_textedit, self.bib_textedit]:
            tii.clear()
            tii.setReadOnly(True)

        self.confirm_review_frame.setVisible(False)

        self.logger.debug('Meta tab cleared.')

        return


    def enableMetaTab(self):

        for kk,vv in self._current_meta_dict.items():
            if kk!='files_l':
                vv.setReadOnly(False)

        for tii in [self.note_textedit, ]:
            tii.setReadOnly(False)

        return


    @pyqtSlot()
    def confirmReviewButtonClicked(self):
        """Confirm meta data of a doc is correct

        This involves:
            * set DocMeta['confirmed']='true'
            * hide self.confirm_reivew_frame.
            * update current doc table
            * remove docid from the Needs Review folder (id='-2')
        """

        docid=self._current_doc

        self.meta_dict[docid]['confirmed']='true'

        self.logger.debug("doc id = %s. meta_dict[docid]['confirmed'] = %s"\
                %(docid, self.meta_dict[docid]['confirmed']))

        self.confirm_review_frame.setVisible(False)
        idx=self.doc_table.currentIndex()
        self.doc_table.model().dataChanged.emit(idx,idx)

        # del doc from needs review folder
        if docid in self.folder_data['-2']:
            self.folder_data['-2'].remove(docid)

        self.loadDocTable(folder=self._current_folder,sortidx=4,sel_row=idx.row())

        return





'''
Defines the main window GUI including the title bar, and database
controls (creation, opening, saving, closing and switching).
Also defines default settings and loading of settings.


MeiTing Trunk

An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import logging
import logging.config
import sqlite3
import pathlib
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, QSettings, QTimer, pyqtSlot
from PyQt5.QtGui import QIcon, QFont, QBrush, QColor
import _MainFrame
from lib import sqlitedb
from lib.widgets import PreferenceDialog, ExportDialog, ThreadRunDialog,\
        ImportDialog, AboutDialog

from main import __version__


# default setting for bibtex export
OMIT_KEYS=[
    'read', 'favourite', 'added', 'confirmed', 'firstNames_l',
    'lastName_l', 'deletionPending', 'folders_l', 'type', 'id',
    'abstract', 'advisor', 'month', 'language', 'confirmed',
    'deletionPending', 'note', 'publicLawNumber', 'sections',
    'reviewedArticle', 'userType', 'shortTitle', 'sourceType',
    'code', 'codeNumber', 'codeSection', 'codeVolume', 'citationKey',
    'day', 'dateAccessed', 'internationalAuthor', 'internationalUserType',
    'internationalTitle', 'internationalNumber', 'genre', 'lastUpdate',
    'legalStatus', 'length', 'medium'
    ]


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self):
        super(MainWindow,self).__init__()

        self.logger=logging.getLogger(__name__)
        self.logger.info('''
##############################################
New session started
##############################################
        ''')
        self.settings=self.initSettings()
        self.is_loaded=False  # is any database opended

        self.main_frame=_MainFrame.MainFrame(self.settings)
        self.main_frame.view_change_sig.connect(self.viewChangeResponse)
        self.setCentralWidget(self.main_frame)

        # put initUI() after main_frame as it's referencing widgets in main_frame
        self.initUI()

        # auto open recent lib
        recent=self.settings.value('file/recent_open',[],str)
        if isinstance(recent,str) and recent=='':
            recent=[]

        self.logger.debug('Recent open list = %s' %recent)

        if self.settings.value('file/auto_open_last',type=int) and len(recent)>0:
            # add a delay, otherwise splash won't show
            QTimer.singleShot(100, lambda: self._openDatabase(recent[-1]))


    def initSettings(self):
        """Load settings file if exists, create new otherwise
        """

        folder_name=os.path.dirname(os.path.abspath(__file__))
        settings_path=os.path.join(folder_name,'settings.ini')

        if not os.path.exists(settings_path):
            settings=QSettings(settings_path,QSettings.IniFormat)

            # fonts for various fields
            settings.setValue('display/fonts/meta_title',
                QFont('Serif', 14, QFont.Bold | QFont.Capitalize))
            settings.setValue('display/fonts/meta_authors',
                QFont('Serif', 11))
            settings.setValue('display/fonts/meta_keywords',
                QFont('Times', 11, QFont.StyleItalic))
            settings.setValue('display/fonts/statusbar',
                QFont('Serif', 10)),
            settings.setValue('display/fonts/doc_table',
                QFont('Serif', 10)),
            settings.setValue('display/fonts/bibtex',
                QFont('Serif', 10)),
            settings.setValue('display/fonts/notes',
                QFont('Serif', 10)),
            settings.setValue('display/fonts/scratch_pad',
                QFont('Serif', 10)),

            # highlight folder containing a doc
            settings.setValue('display/folder/highlight_color_br',
                    QBrush(QColor(200,200,255)))

            settings.setValue('export/bib/omit_fields', OMIT_KEYS)
            settings.setValue('export/bib/path_type', 'absolute')
            settings.setValue('export/ris/path_type', 'absolute')

            # storage recently opened database
            settings.setValue('file/recent_open', [])
            settings.setValue('file/recent_open_num', 2)
            settings.setValue('file/auto_open_last', 1)

            # default storage folder
            storage_folder=os.path.join(str(pathlib.Path.home()),
                    'Documents/MeiTingTrunk')
            settings.setValue('saving/storage_folder', storage_folder)

            # auto save
            settings.setValue('saving/auto_save_min', 5),

            # rename pdf files in storage
            settings.setValue('saving/rename_files', 1)
            settings.setValue('saving/rename_file_replace_space', 1)

            # min score to flag a duplication
            settings.setValue('duplicate_min_score', 60)

            settings.setValue('import/default_add_action', 'Add PDF File')

            # search fields
            settings.setValue('search/search_fields', ['Authors', 'Title',
                'Abstract', 'Keywords', 'Tags', 'Notes', 'Publication'])
            settings.setValue('search/desend_folder', True)

            # view control
            settings.setValue('view/show_widgets', ['Toggle Filter List',
                'Toggle Tab Pane', 'Toggle Meta Tab', 'Toggle Notes Tab',
                'Toggle BibTex Tab', 'Toggle Scratch Pad Tab',
                'Toggle Status bar'])

            settings.sync()
        else:
            settings=QSettings(settings_path,QSettings.IniFormat)

        #---------------Make sure output folder exists---------------
        storage_folder=settings.value('saving/storage_folder')
        self.logger.info('storage_folder=%s' %storage_folder)

        storage_folder=os.path.expanduser(storage_folder)
        if not os.path.exists(storage_folder):
            os.makedirs(storage_folder)

            self.logger.info('Create folder %s' %storage_folder)

        return settings


    def initUI(self):
        """Create menu bar
        """

        self.setWindowTitle('MEI-TING TRUNK %s' %__version__)
        self.setGeometry(100,100,1200,900)    #(x_left,y_top,w,h)
        #self.setWindowIcon(QIcon('img.png'))

        self.menu_bar=self.menuBar()

        #--------------------Files menu--------------------
        self.file_menu=self.menu_bar.addMenu('&File')

        create_database_action=self.file_menu.addAction('Create New Library')
        open_database_action=self.file_menu.addAction('Open Library')
        self.recent_open_menu=self.file_menu.addMenu('Open Recent')
        save_database_action=self.file_menu.addAction('Save Library')
        close_database_action=self.file_menu.addAction('Close Library')
        self.file_menu.addSeparator()
        create_backup_action=self.file_menu.addAction('Create Backup')
        quit_action=self.file_menu.addAction('Quit')

        create_database_action.setIcon(QIcon.fromTheme('document-new'))
        open_database_action.setIcon(QIcon.fromTheme('document-open'))
        self.recent_open_menu.setIcon(QIcon.fromTheme('document-open-recent'))
        save_database_action.setIcon(QIcon.fromTheme('document-save'))
        close_database_action.setIcon(QIcon.fromTheme('call-stop'))
        create_backup_action.setIcon(QIcon.fromTheme('document-send'))
        quit_action.setIcon(QIcon.fromTheme('window-close'))

        create_database_action.setShortcut('Ctrl+Shift+n')
        open_database_action.setShortcut('Ctrl+o')
        save_database_action.setShortcut('Ctrl+s')
        close_database_action.setShortcut('Ctrl+w')
        quit_action.setShortcut('Ctrl+q')

        #---------------Populate open recent---------------
        recent=self.settings.value('file/recent_open',[],str)
        if isinstance(recent,str) and recent=='':
            recent=[]
        recent_num=self.settings.value('file/recent_open_num',type=int)

        if recent and recent_num>0:
            for rii in recent:
                recentii=self.recent_open_menu.addAction(rii)
                recentii.triggered.connect(lambda x,t=rii: self._openDatabase(t))

        #--------------------Edit menu--------------------
        self.edit_menu=self.menu_bar.addMenu('&Edit')
        preference_action=QtWidgets.QAction('Preferences',self)
        preference_action.setIcon(QIcon.fromTheme('preferences-system'))
        preference_action.setShortcut('Ctrl+p')
        self.edit_menu.addAction(preference_action)

        #--------------------View menu--------------------
        self.view_menu=self.menu_bar.addMenu('&View')

        self.view_action_dict={}
        show_widgets=self.settings.value('view/show_widgets',[],str)
        if isinstance(show_widgets,str) and show_widgets=='':
            show_widgets=[]

        for tii in ['Toggle Filter List', None, 'Toggle Tab Pane',
                'Toggle Meta Tab', 'Toggle Notes Tab', 'Toggle BibTex Tab',
                'Toggle Scratch Pad Tab', None, 'Toggle Status bar']:
            if tii is None:
                self.view_menu.addSeparator()
            elif tii=='Toggle Status bar':
                # Status bar is kind of necessary
                continue
            else:
                tii_view_action=self.view_menu.addAction(tii)
                tii_view_action.setCheckable(True)
                if tii in show_widgets:
                    tii_view_action.setChecked(True)
                else:
                    tii_view_action.setChecked(False)
                self.view_action_dict[tii]=tii_view_action

        # remove the box of checkboxes
        self.menu_bar.setStyleSheet('''
        QMenu::indicator {
        }
        ''')

        #--------------------Tools menu--------------------
        self.tool_menu=self.menu_bar.addMenu('&Tools')
        self.import_action=QtWidgets.QAction('&Import', self)
        self.export_action=QtWidgets.QAction('&Export', self)
        self.tool_menu.addAction(self.import_action)
        self.tool_menu.addAction(self.export_action)
        if not self.is_loaded:
            self.export_action.setEnabled(False)

        #--------------------Help menu--------------------
        self.help_menu=self.menu_bar.addMenu('&Help')
        self.help_menu.addAction('&About')

        #-----------------Connect signals-----------------
        create_database_action.triggered.connect(self.createDatabaseTriggered)
        open_database_action.triggered.connect(self.openDatabaseTriggered)
        save_database_action.triggered.connect(self.saveDatabaseTriggered)
        close_database_action.triggered.connect(self.closeDatabaseTriggered)
        preference_action.triggered.connect(self.preferenceTriggered)
        self.import_action.triggered.connect(self.importTriggered)
        self.export_action.triggered.connect(self.exportTriggered)
        self.help_menu.triggered.connect(self.helpMenuTriggered)
        quit_action.triggered.connect(self.close)
        self.view_menu.triggered.connect(self.viewChangeTriggered)

        self.logger.info('Main window UI inited.')

        self.show()

        return


    def closeEvent(self,event):

        if self.is_loaded:
            if len(self.main_frame.changed_folder_ids) +\
                    len(self.main_frame.changed_doc_ids)>0:
                choice=QtWidgets.QMessageBox.question(self, 'Confirm Close',
                        'Save changes and close?',
                        QtWidgets.QMessageBox.Yes |\
                        QtWidgets.QMessageBox.Discard |\
                        QtWidgets.QMessageBox.Cancel)
            else:
                choice=QtWidgets.QMessageBox.question(self, 'Confirm Close',
                        'Close MeiTing Trunk?',
                        QtWidgets.QMessageBox.Yes |\
                        QtWidgets.QMessageBox.Cancel)
        else:
            choice=QtWidgets.QMessageBox.Discard

        if choice==QtWidgets.QMessageBox.Yes:
            self.saveDatabaseTriggered()
            #self.closeDatabaseTriggered(ask=False)
            self.logger.info('settings.sync()')
            self.settings.sync()
            event.accept()
        elif choice==QtWidgets.QMessageBox.Cancel:
            event.ignore()
        elif choice==QtWidgets.QMessageBox.Discard:
            #self.closeDatabaseTriggered(ask=False)
            self.logger.info('settings.sync()')
            self.settings.sync()
            event.accept()

        return


    #######################################################################
    #                           Menu bar actions                           #
    #######################################################################

    @pyqtSlot()
    def createDatabaseTriggered(self):

        # close current if loaded
        if self.is_loaded:
            do=self.closeDatabaseTriggered()
            if not do:
                return

        storage_folder=self.settings.value('saving/storage_folder')
        fname = QtWidgets.QFileDialog.getSaveFileName(self,
                'Create a sqlite file',
                storage_folder,
                "sqlite files (*.sqlite);; All files (*)")[0]

        if fname:

            # make sure has .sqlite ext
            dirname,filename=os.path.split(fname)
            lib_name,ext=os.path.splitext(filename)
            if ext=='':
                filename='%s.sqlite' %lib_name
                fname=os.path.join(dirname,filename)

            def func(jobid,fname):
                try:
                    result=sqlitedb.createNewDatabase(fname)
                    return 0,jobid,result
                except Exception:
                    self.logger.exception('Failed to create new database file')
                    return 1,jobid,None

            # run in separate thread
            td=ThreadRunDialog(func,
                [(0,fname)],
                show_message='Creating new database...',
                max_threads=1,
                get_results=False,
                close_on_finish=True,
                progressbar_style='busy',
                parent=None)

            td.master.all_done_signal.connect(lambda: self._openDatabase(fname))
            td.exec_()


        return


    @pyqtSlot()
    def openDatabaseTriggered(self):

        # close current if loaded
        if self.is_loaded:
            do=self.closeDatabaseTriggered()
            if not do:
                return

        fname = QtWidgets.QFileDialog.getOpenFileName(self,
                'Choose a sqlite file',
                self.settings.value('saving/storage_folder',str),
                "sqlite files (*.sqlite);; All files (*)")[0]

        if fname:
            try:
                self._openDatabase(fname)
            except Exception as e:
                self.logger.exception('Exception in _openDatabase(): %s' %e)

                msg=QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Warning)
                msg.setWindowTitle('Error')
                msg.setText("Oopsi.")
                msg.setInformativeText("Failed to open database file\n    %s" %fname)
                msg.exec_()

            return


    @pyqtSlot(str)
    def _openDatabase(self, fname):
        """Open a sqlite database

        Args:
            fname (str): abs file path to sqlite data file

        """

        # close current if loaded. For open recent calls
        if self.is_loaded:
            do=self.closeDatabaseTriggered()
            if not do:
                return

        if not os.path.exists(fname):
            self.logger.info('Cant find database file %s' %fname)

            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setWindowTitle('Can not find file')
            msg.setText("Can not find target database file.")
            msg.setInformativeText("The requested database file\n    %s\nmay have be deleted, renamed or removed."\
                    %fname)
            msg.exec_()

            # if this is recent, remove it from recent
            recent=self.settings.value('file/recent_open',[],str)
            if isinstance(recent,str) and recent=='':
                recent=[]

            if fname in recent:
                recent.remove(fname)
                self.settings.setValue('file/recent_open', recent)

                self.logger.info('Remove non-exist database file from recent list: %s' %fname)

                for actionii in self.recent_open_menu.findChildren(
                        QtWidgets.QAction):
                    if actionii.text()==fname:
                        self.recent_open_menu.removeAction(actionii)

            return

        self.main_frame.status_bar.showMessage('Opening database...')
        QtWidgets.QApplication.processEvents() # needed?
        # progressbar won't work atm, as the sqlitedb is in the same GUI thread.
        #self.main_frame.progressbar.setVisible(True)
        #self.main_frame.progressbar.setMaximum(0)
        try:
            db = sqlite3.connect(fname)
            self.logger.info('Connected to database: %s' %fname)
        except:
            self.logger.warning('Failed to connect to database %s' %fname)
            return

        self.db=db
        # read and parse data
        meta_dict,folder_data,folder_dict=sqlitedb.readSqlite(db)
        # load data into GUI
        self.main_frame.loadLibTree(db,meta_dict,folder_data,folder_dict)
        #self.main_frame.progressbar.setVisible(False)

        self.is_loaded=True

        # get sqlite file name without ext as library name
        storage_folder,filename=os.path.split(fname)
        lib_name=os.path.splitext(filename)[0]
        lib_folder=os.path.join(storage_folder,lib_name)

        self.current_lib=lib_name
        self.current_lib_folder=lib_folder
        self.settings.setValue('saving/current_lib_folder', self.current_lib_folder)
        self.setWindowTitle('MEI-TING TRUNK %s: %s' %(__version__, lib_name))

        self.logger.info('Get current_lib = %s' %lib_name)
        self.logger.info('Get current_lib_folder = %s' %self.current_lib_folder)

        #-----------Make sure lib folder exists-----------
        if not os.path.exists(self.current_lib_folder):

            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setWindowTitle('Can not find folder')
            msg.setText("Can not find library folder.")
            msg.setInformativeText("The library folder\n    %s\nmay have be deleted, renamed or removed. \nA new folder is created, but the PDF files are missing."\
                    %self.current_lib_folder)
            msg.exec_()

            os.makedirs(self.current_lib_folder)
            self.logger.info('Create lib folder: %s' %self.current_lib_folder)

        lib_collection_folder=os.path.join(self.current_lib_folder,'_collections')
        if not os.path.exists(lib_collection_folder):

            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setWindowTitle('Can not find folder')
            msg.setText("Can not find library folder.")
            msg.setInformativeText("The library folder\n    %s\nmay have be deleted, renamed or removed. \nA new folder is created, but the PDF files are missing."\
                    %lib_collection_folder)
            msg.exec_()
            os.makedirs(lib_collection_folder)
            self.logger.info('Create lib collection folder: %s' %lib_collection_folder)

        self.main_frame.auto_save_timer.start()

        self.logger.info('Start auto save timer.')

        self.import_action.setEnabled(True)
        self.export_action.setEnabled(True)

        # add to recent list
        recent=self.settings.value('file/recent_open',[],str)
        if isinstance(recent,str) and recent=='':
            recent=[]

        if fname in recent:
            recent.remove(fname)

        recent.append(fname)
        recentii=self.recent_open_menu.addAction(fname)
        recentii.triggered.connect(lambda x,t=fname: self._openDatabase(t))

        # pop 1st from recent list if exceeding limit
        recent_open_num=self.settings.value('file/recent_open_num',type=int)
        if len(recent)>recent_open_num:
            recent.pop(0)

        self.settings.setValue('file/recent_open',recent)

        return


    @pyqtSlot()
    def saveDatabaseTriggered(self):
        self.main_frame.saveToDatabase()
        return


    @pyqtSlot()
    def closeDatabaseTriggered(self,ask=True):

        if ask:
            choice=QtWidgets.QMessageBox.question(self, 'Confirm Close',
                    'Save and close current library?',
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        if not ask or (ask and choice==QtWidgets.QMessageBox.Yes):
            self.main_frame.clearData()
            self.is_loaded=False

            self.import_action.setEnabled(False)
            self.export_action.setEnabled(False)

            self.current_lib=None
            self.current_lib_folder=None
            self.settings.setValue('saving/current_lib_folder','')
            self.setWindowTitle('MEI-TING TRUNK %s' %__version__)
            self.main_frame.auto_save_timer.stop()
            self.main_frame.status_bar.clearMessage()

            self.logger.info('Stop auto save timer.')
            self.logger.info('Database closed.')

            return True
        else:
            return False


    @pyqtSlot(QtWidgets.QAction)
    def helpMenuTriggered(self,action):
        self.logger.info('action=%s, action.text()=%s' %(action,action.text()))
        dia=AboutDialog(__version__, self)
        dia.exec_()
        return


    @pyqtSlot()
    def preferenceTriggered(self):

        diag=PreferenceDialog(self.settings,parent=self)
        diag.exec_()

        return


    @pyqtSlot()
    def importTriggered(self):

        diag=ImportDialog(self.settings,parent=self)
        # for open newly imported database
        diag.open_lib_signal.connect(self._openDatabase)
        diag.exec_()

        return


    @pyqtSlot()
    def exportTriggered(self):

        diag=ExportDialog(self.settings,parent=self)
        diag.exec_()

        return


    @pyqtSlot(QtWidgets.QAction)
    def viewChangeTriggered(self, action):
        """Change widget visibility in response to View menu actions

        Args:
            action (QAction): QAction in View menu
        """

        action_text=action.text()

        self.logger.info('View change action = %s' %action_text)

        if action_text=='Toggle Filter List':
            self.main_frame.foldFilterButtonClicked()
        elif action_text=='Toggle Tab Pane':
            self.main_frame.foldTabButtonClicked()
        elif action_text=='Toggle Status bar':
            self.main_frame.statusbarViewChange()
        else:
            self.main_frame.metaTabViewChange(action_text)

        return


    @pyqtSlot(str,bool)
    def viewChangeResponse(self, view_name, state):
        """Store widget visibility states in response to hide/show buttons

        Args:
            view_name (str): name of a widget view action, also the key
                             in view_action_dict. See the View menu creation
                             section in initUI().

            state (bool): Visibility state of a widget.
        """

        self.view_action_dict[view_name].setChecked(state)




'''
Widgets for search computation and result display.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from collections import OrderedDict
import logging
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QSize, QRegExp, QRect
from PyQt5.QtGui import QBrush, QColor, QFont, \
        QSyntaxHighlighter, QTextCharFormat
#from PyQt5.QtWidgets import QDialogButtonBox
from .. import sqlitefts
from ..tools import iterTreeWidgetItems


LOGGER=logging.getLogger(__name__)



class DummyWidget(QtWidgets.QWidget):
    '''A dummy widget to hold the fold button, text label and textedit for a
    search result detail row
    '''

    resize_sig=pyqtSignal(QSize)

    def __init__(self,parent=None):
        super(DummyWidget,self).__init__(parent)

    def resizeEvent(self,e):
        super(DummyWidget,self).resizeEvent(e)
        self.resize_sig.emit(self.sizeHint())



class HighLighter(QSyntaxHighlighter):
    '''Highlight the search term in textedit'''

    def __init__(self,match_words,parent=None):
        super(HighLighter,self).__init__(parent)

        brush=QBrush(Qt.yellow, Qt.SolidPattern)
        keyword=QTextCharFormat()
        keyword.setBackground(brush)
        keyword.setFontWeight(QFont.Bold)

        self.highlightingRules = [(QRegExp(key), keyword) for key in match_words]

    def highlightBlock(self, text):

        for pattern, tformat in self.highlightingRules:
            expression=QRegExp(pattern)
            index=expression.indexIn(text)

            while index >= 0:
              length = expression.matchedLength()
              self.setFormat(index, length, tformat)
              index = expression.indexIn(text, index + length)

        self.setCurrentBlockState(0)



class AdjustableTextEdit(QtWidgets.QTextEdit):
    '''This modified QTextEdit doesn't show scroll bar, but adjusts height
    acorrding to contents and width.
    '''

    td_size_sig=pyqtSignal(QSize)
    def __init__(self,parent=None):

        super(AdjustableTextEdit,self).__init__(parent)

        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.textChanged.connect(self.resizeTextEdit)
        self.document().documentLayout().documentSizeChanged.connect(
                self.resizeTextEdit)


    def resizeTextEdit(self):
        '''Set textedit height according to contents'''

        docheight=self.document().size().height()
        margin=self.document().documentMargin()
        self.setMinimumHeight(docheight+2*margin)
        self.setMaximumHeight(docheight+2*margin)

        return


    def resizeEvent(self,e):
        '''Send the current size via signal'''

        super(AdjustableTextEdit,self).resizeEvent(e)
        self.td_size_sig.emit(QSize(self.sizeHint().width(),
            self.maximumHeight()))

        return


    def setHighlightText(self, text_list):
        '''Highlight the search term'''

        if not isinstance(text_list, (tuple, list)):
            text_list=[text_list,]
        HighLighter(text_list, self)

        return


class AdjustableTextEditWithFold(AdjustableTextEdit):
    '''Adjustable textedit with a fold button to fold/expand long texts'''

    fold_size_sig=pyqtSignal(QSize)
    def __init__(self,parent=None):
        super(AdjustableTextEditWithFold,self).__init__(parent)

        self.is_fold=True
        self.fold_above_nl=2

        self.fold_button=QtWidgets.QToolButton()
        self.fold_button.setArrowType(Qt.DownArrow)
        font_height=self.fontMetrics().height()
        self.fold_button.setFixedWidth(int(font_height))
        self.fold_button.setFixedHeight(int(font_height))
        self.fold_button.clicked.connect(self.toggleFold)
        self.fold_button.setStyleSheet('''
        QToolButton {
            border: 1px solid rgb(190,190,190);
            background-color: rgb(190,190,190);
            border-radius: %dpx;
            color: white;
            }

        QToolButton:pressed {
            border-style: inset;
            }
        ''' %(int(font_height/2))
        )

    def getNumberOfLines(self):

        fm=self.fontMetrics()
        doc=self.document()
        docheight=doc.size().height()
        margin=doc.documentMargin()
        nlines=(docheight-2*margin)/fm.height()

        return nlines


    def resizeTextEdit(self):
        '''Show/hide fold button according to number of lines in text'''

        if self.getNumberOfLines()<self.fold_above_nl:
            self.fold_button.setVisible(False)
        else:
            self.fold_button.setVisible(True)
        if self.is_fold:
            self.foldText()
        else:
            self.unfoldText()

        return


    def toggleFold(self):

        self.unfoldText() if self.is_fold else self.foldText()
        self.fold_size_sig.emit(QSize(self.sizeHint().width(),
            self.maximumHeight()))

        return


    def foldText(self):

        nlines=self.getNumberOfLines()
        if nlines>=self.fold_above_nl:
            fontheight=self.fontMetrics().height()
            margin=self.document().documentMargin()
            self.setMinimumHeight(fontheight+2*margin)
            self.setMaximumHeight(fontheight+2*margin)
            self.is_fold=True
            self.fold_button.setArrowType(Qt.RightArrow)

        return


    def unfoldText(self):

        docheight=self.document().size().height()
        margin=self.document().documentMargin()
        self.setMinimumHeight(docheight+2*margin)
        self.setMaximumHeight(docheight+2*margin)
        self.is_fold=False
        self.fold_button.setArrowType(Qt.DownArrow)

        return



class BorderItemDelegate(QtWidgets.QStyledItemDelegate):
    '''To show borders in treewidget item.
    NOT IN USE
    '''

    def __init__(self, parent, borderRole):
        super(BorderItemDelegate, self).__init__(parent)
        self.borderRole = borderRole

    def sizeHint(self, option, index):
        size = super(BorderItemDelegate, self).sizeHint(option, index)
        pen = index.data(self.borderRole)
        if pen is not None:
            # Make some room for the border
            # When width is 0, it is a cosmetic pen which
            # will be 1 pixel anyways, so set it to 1
            width = max(pen.width(), 1)
            size = size + QSize(2 * width, 2 * width)
        return size

    def paint(self, painter, option, index):
        pen = index.data(self.borderRole)
        # copy the rect for later...
        rect = QRect(option.rect)
        if pen is not None:
            width = max(pen.width(), 1)
            # ...and remove the extra room we added in sizeHint...
            option.rect.adjust(width, width, -width, -width)

        # ...before painting with the base class method...
        super(BorderItemDelegate, self).paint(painter, option, index)

        # ...then paint the borders
        if pen is not None:
            painter.save()
            # The pen is drawn centered on the rectangle lines
            # with pen.width()/2 width on each side of these lines.
            # So, rather than shifting the drawing of pen.width()/2
            # we double the pen width and clip the part that would
            # go outside the rect.
            painter.setClipRect(rect, Qt.ReplaceClip);
            pen.setWidth(2 * width)
            painter.setPen(pen)
            painter.drawRect(rect)
            painter.restore()



class SearchResFrame(QtWidgets.QScrollArea):

    search_done_sig=pyqtSignal()
    #hide_doc_sig=pyqtSignal()
    create_folder_sig=pyqtSignal(str,list)  # search_text, docids
    MyBorderRole = Qt.UserRole + 1  # not in use

    def __init__(self,settings,parent=None):
        '''Handles text searching and result display

        Args:
            parent (QWidget): parent widget.
            settings (QSettings): application settings. See _MainWindow.py
        '''
        super(SearchResFrame,self).__init__(parent=parent)

        self.settings=settings
        self.parent=parent

        frame=QtWidgets.QWidget()
        self.setWidgetResizable(True)
        self.setWidget(frame)
        va=QtWidgets.QVBoxLayout(self)

        #----------------Create clear frame----------------
        va.addWidget(self.createClearSearchFrame())

        #----------------Create treewidget----------------
        self.tree=QtWidgets.QTreeWidget(self)
        self.tree.setColumnCount(6)
        # put doc id in column 5. This is later used to identify selected docs.
        self.tree.setColumnHidden(5,True)

        # column widths
        headers=OrderedDict([
            ('Fold all', 70),
            ('Authors', 150),
            ('Title', 300),
            ('Publication', 100),
            ('Year', 50),
            ('id', 0)
            ])

        self.tree.setHeaderLabels(headers.keys())
        for ii,kk in enumerate(headers.keys()):
            self.tree.setColumnWidth(ii, headers[kk])

        self.tree.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.tree.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.tree.header().setStretchLastSection(True)
        self.tree.header().setSectionResizeMode(
                QtWidgets.QHeaderView.Interactive)
        self.tree.header().setSectionsClickable(True)
        self.tree.header().sectionClicked.connect(self.headerSectionClicked)
        self.tree.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tree.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.tree.setDragDropMode(QtWidgets.QAbstractItemView.NoDragDrop)
        #self.tree.setContextMenuPolicy(Qt.CustomContextMenu)
        self.tree.itemSelectionChanged.connect(self.changeBGColor)

        self.is_all_fold=False

        self.noMatchLabel=QtWidgets.QLabel('No match found.')
        self.noMatchLabel.setVisible(False)
        va.addWidget(self.noMatchLabel)
        va.addWidget(self.tree)

        frame.setLayout(va)


    def createClearSearchFrame(self):
        '''Create a header frame showing search term, results count and control
        buttons'''

        frame=QtWidgets.QFrame()
        frame.setStyleSheet('background: rgb(235,225,190)')
        ha=QtWidgets.QHBoxLayout()

        # create folder button
        self.create_folder_button=QtWidgets.QToolButton(self)
        self.create_folder_button.setText('Create Folder From Selection')
        self.create_folder_button.clicked.connect(self.createFolder)

        # hide doc table button
        #self.hide_doc_table_button=QtWidgets.QToolButton(self)
        #self.hide_doc_table_button.setText('Hide Document Table')
        #self.hide_doc_table_button.clicked.connect(lambda: self.hide_doc_sig.emit())

        # clear button
        self.clear_searchres_button=QtWidgets.QToolButton(self)
        self.clear_searchres_button.setText('Close')

        self.label=QtWidgets.QLabel('Search results')
        ha.addWidget(self.label)
        ha.addWidget(self.create_folder_button)
        #ha.addWidget(self.hide_doc_table_button)
        ha.addWidget(self.clear_searchres_button)

        frame.setLayout(ha)

        return frame


    def addFieldRows(self, parent, fields, meta, search_text):
        '''Add detail rows to a matched doc

        Args:
            parent (QTreeWidgetItem): the header item showing a matched doc.
            fields (list): list of matched field names, including 'authors',
                           'title', 'keywords' etc..
            meta (DocMeta): meta data dict for the matched doc.
            search_text (str): searched text.

        A dummy container widget is created, inside which a grid layout is used
        to hold a number of rows, each being a matched field, e.g. 'authors',
        'title', etc.. Each row consists of a fold button to fold long texts,
        a label showing the field name, and a textedit showing the field value,
        with the <search_text> highlighted.

        The dummy container is then put in a QTreeWidgetItem by calling
        setItemWidget(). The item is added as a child of <parent>, which
        shows the matched doc in similar format as the doc table.
        '''

        font=self.settings.value('display/fonts/doc_table',QFont)

        item=QtWidgets.QTreeWidgetItem()
        parent.addChild(item)
        item.setFirstColumnSpanned(True)

        frame=DummyWidget()
        grid=QtWidgets.QGridLayout(frame)
        crow=grid.rowCount()

        for fii in fields:
            if fii=='authors':
                textii=', '.join(meta['authors_l'])
            elif fii=='title':
                textii=meta['title']
            elif fii=='publication':
                textii=meta['publication']
            elif fii=='keywords':
                textii=', '.join(meta['keywords_l'])
            elif fii=='tag':
                textii=', '.join(meta['tags_l'])
            elif fii=='abstract':
                textii=meta['abstract']
            elif fii=='note':
                textii=meta['notes']
            else:
                LOGGER.warning('Wrong field given %s' %fii)
                raise Exception("Exception")

            labelii=QtWidgets.QLabel('%s: ' %fii)
            text_editii=AdjustableTextEditWithFold()
            text_editii.setFont(font)
            text_editii.setText(textii)
            text_editii.setHighlightText(search_text)
            grid.addWidget(text_editii.fold_button,crow,0)
            grid.addWidget(labelii,crow,1)
            grid.addWidget(text_editii,crow,2)

            text_editii.fold_size_sig.connect(lambda x: frame.resize(
                frame.sizeHint()))
            crow+=1

        self.tree.setItemWidget(item,0,frame)
        # add doc id to column 5
        item.setText(5,str(meta['id']))
        frame.resize_sig.connect(lambda size: (item.setSizeHint(0,size),
            self.tree.model().layoutChanged.emit()))

        return


    def changeBGColor(self):
        '''Change background color of QTreeWidgetItem

        This makes the header row of a doc, and all detail rows (child items
        of the header row) appeared to be selected together, by giving them
        a highlight color. Selecting either the header row or any child row
        makes them all highlighted.
        '''

        sel_rows=self.tree.selectedItems()
        hlcolor=self.tree.palette().highlight().color().name()

        docids=[itemii.data(5,0) for itemii in sel_rows]
        root=self.tree.invisibleRootItem()

        # set all widgetitems with selected docids selected.
        for item in iterTreeWidgetItems(self.tree, root):
            idii=item.data(5,0)
            if idii in docids:
                # I thought this would results in endless loop, but it doesn't!
                item.setSelected(True)

            # child detail rows has item widget
            wii=self.tree.itemWidget(item,0)
            if wii:
                if idii in docids:
                    wii.setStyleSheet('background-color: %s;' %hlcolor)
                else:
                    wii.setStyleSheet('')

        return


    def search(self, db, text, field_list, folderid, meta_dict, desend):
        """Start search

        Args:
            db (sqlite connection): sqlite connection.
            text (str): searched term.
            field_list (list): list of field names to search. Including
                'Authors', 'Title', etc.. The complete list see _MainFrame.py.
            folderid (str): id the folder. Search is done within docs in this
                 folder, also controled by <desend>.
            meta_dict (dict): dict of all meta data in the library.
            desend (bool): whether to walk down the folder tree and include
                 subfolders of folder <folderid>.

        """

        self.tree.clear()
        self.setVisible(True)
        self.meta_dict=meta_dict
        self.search_text=text
        self.desend=desend
        LOGGER.info('search text = %s. is desend = %s' %(text, desend))

        search_res=sqlitefts.searchMultipleLike2(db, text, field_list,
                folderid, desend)
        self.label.setText('%d searches results related to "%s"'\
                %(len(search_res),text))
        self.addResultToTree(text, search_res)

        return


    def addResultToTree(self, search_text, search_res):
        '''Add search results to a QTreeWidget for display

        Args:
            search_text (str): searched term.
            search_res (list): list of doc ids matching search, together with
                the field names where the match is found. E.g.
                        [(1, 'authors,title'),
                         (10, 'title,keywords'),
                         (214, 'abstract,tag'),
                         ...
                        ]
        '''

        def createEntry(docid, gid):
            '''Create an entry for a matched doc

            Args:
                docid (int): id of matched doc.
                gid (int): counter of the current doc.
            '''

            meta=self.meta_dict[docid]
            item=QtWidgets.QTreeWidgetItem([
                gid,
                ', '.join(meta['authors_l']),
                meta['title'],
                meta['publication'],
                str(meta['year']),
                str(docid)
                ])
            return item

        #-------------------If no match-------------------
        if len(search_res)==0:
            self.noMatchLabel.setVisible(True)
            LOGGER.info('Not result found.')
            return

        self.noMatchLabel.setVisible(False)

        #-------------Create entries for docs-------------
        for ii,recii in enumerate(search_res):
            docii, fieldii=recii
            fieldii=list(set(fieldii.split(','))) # str to list
            itemii=createEntry(docii, str(ii+1))
            self.tree.addTopLevelItem(itemii)

            # add group members
            self.addFieldRows(itemii, fieldii, self.meta_dict[docii],
                    search_text)

        self.tree.expandAll()

        # highlight header rows
        hi_color=self.settings.value('display/folder/highlight_color_br',
                QBrush).color().name()

        self.tree.setStyleSheet('''
        QTreeWidget::item:has-children:!selected {
        border-left: 1px solid black;
        border-bottom: 1px solid black;
        background-color: %s;
        }
        ''' %hi_color)

        self.search_done_sig.emit() # trigger main_frame.status_bar.clearMessage()

        return


    @pyqtSlot()
    def createFolder(self):
        '''Send signal to create a folder for selected search results'''

        sel_rows=self.tree.selectedItems()
        if len(sel_rows)>0:
            docids=[]
            for ii in sel_rows:
                docids.append(int(ii.data(5,0)))
            docids=list(set(docids))
            LOGGER.info('Selected docids=%s.' %docids)
            self.create_folder_sig.emit(self.search_text, docids)

        return


    @pyqtSlot(int)
    def headerSectionClicked(self,idx):
        '''Expand/fold all docs on clicking header 0'''

        if idx==0:
            if self.is_all_fold:
                LOGGER.debug('fold all')

                for ii in range(self.tree.topLevelItemCount()):
                    itemii=self.tree.topLevelItem(ii)
                    self.tree.expandItem(itemii)
                self.is_all_fold=False
                self.tree.setHeaderLabels(['Fold all', 'Authors', 'Title',
                    'Publication', 'Year', 'id'])
            else:
                LOGGER.debug('expand all')

                for ii in range(self.tree.topLevelItemCount()):
                    itemii=self.tree.topLevelItem(ii)
                    self.tree.collapseItem(itemii)
                self.is_all_fold=True
                self.tree.setHeaderLabels(['Unfold all', 'Authors', 'Title',
                    'Publication', 'Year', 'id'])




from queue import Queue
import logging
from PyQt5 import QtWidgets
from PyQt5.QtCore import QObject, QThread, Qt, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QDialogButtonBox


LOGGER=logging.getLogger('default_logger')



class Worker(QObject):

    worker_jobdone_signal=pyqtSignal(int) # jobid

    def __init__(self, id, func, jobqueue, outqueue):
        super(Worker,self).__init__()
        '''Worker used for separate thread.

        NOTE: args for func is assumed to have the format (jobid,jobargs)
        return of func is assumed to have the format (return_code, jobid,\
                other_results)
        '''

        self.id=id
        self.func=func
        self.jobqueue=jobqueue
        self.outqueue=outqueue
        self.abort=False

    @pyqtSlot()
    def processJob(self):
        while self.jobqueue.qsize():
            QtWidgets.QApplication.processEvents()
            if self.abort:
                return

            args=self.jobqueue.get()
            jobid=args[0]
            try:
                rec=self.func(*args)
                print('# <Worker>: Job %d done. Remaining queue size: %d.'\
                    %(jobid, self.jobqueue.qsize()))
            except:
                rec=(1,jobid,None)
                print('# <Worker>: Job %d failed. Remaining queue size: %d.'\
                    %(jobid, self.jobqueue.qsize()))
            self.jobqueue.task_done()
            self.outqueue.put(rec)
            self.worker_jobdone_signal.emit(jobid)

        return



class Master(QObject):

    all_done_signal=pyqtSignal()
    donejobs_count_signal=pyqtSignal(int) # NO. of finshed jobs

    def __init__(self, func, joblist, max_threads=4):
        super(Master,self).__init__()

        self.func=func
        self.joblist=joblist
        self.max_threads=max_threads
        self.all_done_signal.connect(self.onAllJobsDone)

    def run(self):

        self.threads=[]
        self.results=[]
        n_threads=min(self.max_threads,len(self.joblist))
        self.finished=0
        self.finished_jobs=[]

        self.jobqueue=Queue()
        self.outqueue=Queue()
        # populate job queue
        for ii,jobii in enumerate(self.joblist):
            self.jobqueue.put(jobii)

        # start worker threads
        for ii in range(n_threads):
            print('# <run>: create thread',ii)
            tii=QThread()
            wii=Worker(ii,self.func,self.jobqueue,self.outqueue)
            self.threads.append((tii,wii)) # need to keep record of both!

            wii.moveToThread(tii)
            wii.worker_jobdone_signal.connect(self.countJobDone)
            tii.started.connect(wii.processJob)
            tii.start()

        return


    @pyqtSlot(int)
    def countJobDone(self,jobid):
        self.finished+=1
        self.finished_jobs.append(jobid)
        self.donejobs_count_signal.emit(self.finished)
        while self.outqueue.qsize():
            resii=self.outqueue.get()
            self.results.append(resii)

        print('# <countJobDone>: Finished job id=%d. Finished jobs=%d'\
                %(jobid, self.finished))

        if self.finished==len(self.joblist):
            self.all_done_signal.emit()

        return


    @pyqtSlot()
    def onAllJobsDone(self):
        while self.outqueue.qsize():
            resii=self.outqueue.get()
            self.results.append(resii)

        for tii,wii in self.threads:
            tii.quit()
            tii.wait()
        return

    @pyqtSlot()
    def abortJobs(self):
        for tii,wii in self.threads:
            wii.abort=True
            tii.quit()
            tii.wait()



class ThreadRunDialog(QtWidgets.QDialog):

    def __init__(self,func,joblist,show_message='',max_threads=3,
            get_results=False,close_on_finish=True,parent=None):
        super(ThreadRunDialog,self).__init__(parent=parent)

        self.func=func
        self.joblist=joblist
        self.show_message=show_message
        self.max_threads=max_threads
        self.get_results=get_results
        self.close_on_finish=close_on_finish
        self.parent=parent

        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle('Processing ...')
        self.resize(500,100)

        va=QtWidgets.QVBoxLayout(self)

        va.addWidget(QtWidgets.QLabel(show_message))

        self.progressbar=QtWidgets.QProgressBar(self)
        self.progressbar.setMaximum(len(joblist))
        self.progressbar.setValue(0)

        self.buttons=QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, self)

        va.addWidget(self.progressbar)
        va.addWidget(self.buttons)

        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.abortJobs)

        self.master=Master(func,joblist,self.max_threads)
        self.master.donejobs_count_signal.connect(self.updatePB)
        self.ok_button=self.buttons.button(QDialogButtonBox.Ok)
        self.ok_button.setEnabled(False)
        self.cancel_button=self.buttons.button(QDialogButtonBox.Cancel)
        self.master.all_done_signal.connect(self.allJobsDone)

        self.master.run()
        self.exec_()


    @pyqtSlot()
    def allJobsDone(self):
        self.cancel_button.setEnabled(False)
        self.ok_button.setEnabled(True)
        if self.close_on_finish:
            self.accept()
        return

    @pyqtSlot(int)
    def updatePB(self,value):
        self.progressbar.setValue(value)
        return

    @pyqtSlot()
    def abortJobs(self):
        self.master.abortJobs()
        if self.get_results:
            self.results=self.master.results
        self.reject()
        return

    @pyqtSlot()
    def accept(self):
        print('# <accept>: get result?',self.get_results)
        if self.get_results:
            self.results=self.master.results
            print('# <accept>: self.results',self.results)
        super(ThreadRunDialog,self).accept()
        return





'''
Widgets for the folder tree: a QItemDelegate and a QTreeWidget


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import logging
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QRegExpValidator

LOGGER=logging.getLogger(__name__)



class TreeWidgetDelegate(QtWidgets.QItemDelegate):
    def __init__(self, parent=None):
        QtWidgets.QItemDelegate.__init__(self, parent=parent)

    def createEditor(self, parent, option, index):
        '''To disable typing in invalid chars in the folder name'''

        editor = QtWidgets.QLineEdit(parent)
        reg=QtCore.QRegExp('[A-z0-9\[\]_-\s]+')
        vd=QRegExpValidator(reg)
        editor.setValidator(vd)

        return editor

    '''
    def eventFilter(self, editor, event):
        # NOTE: possibly move the name conflict check inside
        print('# <eventFilter>: editor:', editor, 'event', event)


        if event.type()==QEvent.KeyPress:
            if event.key()==Qt.Key_Enter or event.key()==Qt.Key_Return:
                print('# <eventFilter>: Keypress', event.key())

                if editor.text()=='aaa':
                    return False
                self.commitData.emit(editor)
                self.closeEditor.emit(editor,0)
                return True
            else:
                return False
        else:
            return False
    '''


class MyTreeWidget(QtWidgets.QTreeWidget):

    folder_move_signal=pyqtSignal(str, str) # move_folder_id, new_parent_id
    folder_del_signal=pyqtSignal(QtWidgets.QTreeWidgetItem,  # move item
            QtWidgets.QTreeWidgetItem,  # newparent
            bool  # ask
            )
    add_doc_to_folder_signal=pyqtSignal(int, str)  # docid, folderid

    def __init__(self,parent=None):
        self.parent=parent
        super(MyTreeWidget,self).__init__(parent=parent)

        self.setDropIndicatorShown(True)
        self.setHeaderHidden(True)
        # column1: folder name, column2: folder id
        self.setColumnCount(2)
        self.hideColumn(1)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        # make horziontal scroll bar appear
        self.header().setStretchLastSection(False)
        self.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)

        delegate=TreeWidgetDelegate()
        self.setItemDelegate(delegate)


    def commitData(self,widget):
        '''Things to do after editing folder name

        itemChanged signal is connected to addNewFolderToDict to update the
        new folder name to storage. Then disconnect immediately.
        '''

        self.itemChanged.connect(self.parent.addNewFolderToDict, Qt.QueuedConnection)
        super(MyTreeWidget,self).commitData(widget)
        recs=self.receivers(self.itemChanged)

        LOGGER.debug('receivers of itemChanged signal = %s' %recs)
        if recs>0:
            self.itemChanged.disconnect()

        return


    def selectedIndexes(self):
        # to fix the issue of missing columns when setting DragDrop mode.
        return self.selectionModel().selectedIndexes()


    def startDrag(self,actions):

        move_item=self.selectedItems()[0]
        LOGGER.debug('move_item.data(0,0) = %s, move_item.data(1,0) = %s'\
                %(move_item.data(0,0), move_item.data(1,0)))

        self._move_item=move_item
        # abort if item is system folders
        if move_item.data(1,0) in ['-1','-2','-3']:
            return

        super(MyTreeWidget,self).startDrag(actions)


    def dragEnterEvent(self,event):

        #--------Deny dragging to all, needs review--------
        pos=event.pos()
        newparent=self.itemAt(pos)
        if newparent is not None and newparent.data(1,0) in ['-2','-1']:
            event.ignore()
            return

        mime_data=event.mimeData()
        LOGGER.debug('event.mimeData() = %s. mime_data.formats() = %s'\
                %(mime_data, mime_data.formats()))

        #----------------Dragging a doc row----------------
        if mime_data.hasFormat('doc_table_item'):
            trashed_folders=self.parent._trashed_folder_ids
            current_item=self.selectedItems()
            if current_item:
                current_item=current_item[0]

            LOGGER.debug('folderid of current item = %s' %current_item.data(1,0))
            LOGGER.debug('folderid of target item = %s' %newparent.data(1,0))

            if newparent is not None and current_item is not None and\
                    newparent.data(1,0) not in trashed_folders and\
                    current_item.data(1,0) in trashed_folders:
                event.setDropAction(Qt.MoveAction)

                LOGGER.warning('Set drop action to Qt.MoveAction. Doesnt seem to work.')
            else:
                event.setDropAction(Qt.CopyAction)

            event.acceptProposedAction()

        #----------------Dragging a folder----------------
        elif mime_data.hasFormat('application/x-qabstractitemmodeldatalist'):
            event.setDropAction(Qt.MoveAction)
            event.acceptProposedAction()
        else:
            event.ignore()

        return

    def dragMoveEvent(self,event):

        #--------Deny dragging to all, needs review--------
        pos=event.pos()
        newparent=self.itemAt(pos)
        if newparent is not None and newparent.data(1,0) in ['-2','-1']:
            event.ignore()
            return

        mime_data=event.mimeData()

        #----------------Dragging a doc----------------
        if mime_data.hasFormat('doc_table_item'):
            trashed_folders=self.parent._trashed_folder_ids
            current_item=self.selectedItems()
            if current_item:
                current_item=current_item[0]

            LOGGER.debug('folderid of current item = %s' %current_item.data(1,0))
            LOGGER.debug('folderid of target item = %s' %newparent.data(1,0))

            if newparent is not None and current_item is not None and\
                    newparent.data(1,0) not in trashed_folders and\
                    current_item.data(1,0) in trashed_folders:
                event.setDropAction(Qt.MoveAction)
                LOGGER.warning('Set drop action to Qt.MoveAction. Doesnt seem to work.')
            else:
                event.setDropAction(Qt.CopyAction)

            event.acceptProposedAction()

        #----------------Dragging a folder----------------
        elif mime_data.hasFormat('application/x-qabstractitemmodeldatalist'):

            # deny droping to self
            pos=event.pos()
            newparent=self.itemAt(pos)
            if newparent==self._move_item:
                event.ignore()
                return

            event.setDropAction(Qt.MoveAction)
            event.acceptProposedAction()

            # need this to make dropIndicatorPosition work
            super(MyTreeWidget,self).dragMoveEvent(event)
        else:
            event.ignore()

        return


    def dropEvent(self,event):

        mime_data=event.mimeData()

        #------------------dropping a doc------------------
        if mime_data.hasFormat('doc_table_item'):
            # decode byte to str
            dropped_docid=mime_data.data('doc_table_item').data().decode('ascii')
            dropped_docid=int(dropped_docid)
            pos=event.pos()
            newparent=self.itemAt(pos)
            parentidx=self.indexFromItem(newparent)

            LOGGER.debug('docid = %s. prarentid = %s.' %(dropped_docid,
                newparent.data(1,0)))

            if newparent.data(1,0) not in ['', '-2', '-1']:
                LOGGER.info('Doc drop valid. Emitting add_doc_to_folder_signal.')
                self.add_doc_to_folder_signal.emit(dropped_docid, newparent.data(1,0))

            return

        #----------------dropping a folder----------------
        elif mime_data.hasFormat('application/x-qabstractitemmodeldatalist'):
            event.setDropAction(Qt.MoveAction)

            if self._move_item.data(1,0) in ['-1','-2','-3']:
                return

            pos=event.pos()
            newparent=self.itemAt(pos)
            parentidx=self.indexFromItem(newparent)
            indicatorpos=self.dropIndicatorPosition()

            LOGGER.debug('parentidx.row() = %s. newparent=data(0,0) = %s'\
                    %(parentidx.row(), newparent.data(0,0)))
            LOGGER.debug('dropIndicatorPosition = %s' %indicatorpos)

            #-----------------dropping on item-----------------
            if indicatorpos==0:

                if newparent.data(1,0) in ['-1', '-2']:
                    event.ignore()
                    return

                #----------------dropping to trash----------------
                elif newparent.data(1,0) in ['-3']+self.parent._trashed_folder_ids:

                    LOGGER.info('Dropping to trash a folder. Emitting folder_del_signal')
                    self.folder_del_signal.emit(self._move_item,newparent,True)

                    return

                # get children
                children=[newparent.child(ii) for ii in range(newparent.childCount())]
                children_names=[ii.data(0,0) for ii in children]
                LOGGER.debug('Got children names = %s' %children_names)

                # check name conflict
                if self._move_item.data(0,0) in children_names:

                    LOGGER.info('Found folder name conflict. Folder name = %s'\
                            %self._move_item.data(0,0))
                    event.ignore()
                    msg=QtWidgets.QMessageBox()
                    msg.setIcon(QtWidgets.QMessageBox.Critical)
                    msg.setWindowTitle('Name conflict')
                    msg.setText('Move cancelled due to name conflict.')
                    msg.setInformativeText('Folder name\n\t%s\nconflicts with another folder in the target folder.\nPlease rename before moving.' %self._move_item.data(0,0))
                    msg.exec_()

                    return
                else:
                    # change folder parent
                    event.setDropAction(Qt.MoveAction)
                    LOGGER.info('Folder drop valid. Emit folder_move_signal')

                    self.folder_move_signal.emit(self._move_item.data(1,0),\
                            newparent.data(1,0))
                    super(MyTreeWidget,self).dropEvent(event)

                    return

            #--------------dropping above an item--------------
            elif indicatorpos==1:
                if parentidx.row()<=3:
                    event.ignore()
                    return
                else:
                    grandparent=newparent.parent()
                    if grandparent is None:
                        grandparentid='-1'
                    else:
                        grandparentid=grandparent.data(1,0)

                    event.setDropAction(Qt.MoveAction)
                    LOGGER.debug('grandparentid = %s' %grandparentid)
                    LOGGER.info('Folder drop valid. Emitting folder_move_signal.')

                    self.folder_move_signal.emit(self._move_item.data(1,0),\
                            grandparentid)
                    super(MyTreeWidget,self).dropEvent(event)

                    return

            #--------------dropping below an item--------------
            elif indicatorpos==2:
                if parentidx.row()<=2:
                    event.ignore()
                    return
                else:
                    grandparent=newparent.parent()
                    if grandparent is None:
                        grandparentid='-1'
                    else:
                        grandparentid=grandparent.data(1,0)

                    event.setDropAction(Qt.MoveAction)
                    LOGGER.debug('grandparentid = %s' %grandparentid)
                    LOGGER.info('Folder drop valid. Emitting folder_move_signal.')

                    self.folder_move_signal.emit(self._move_item.data(1,0),\
                            grandparentid)
                    super(MyTreeWidget,self).dropEvent(event)

                    return





from .folder_tree import TreeWidgetDelegate, MyTreeWidget
from .doc_table import TableModel, MyHeaderView
from .meta_tab import MetaTabScroll, NoteTextEdit
from .preference_dialog import PreferenceDialog
from .export_dialog import ExportDialog
from .duplicate_frame import CheckDuplicateFrame
from .threadrun_dialog import ThreadRunDialog, Master
from .fail_dialog import FailDialog
from .search_res_frame import SearchResFrame
from .import_dialog import ImportDialog
from .create_new_dialog import CreateNewLibDialog
from .about_dialog import AboutDialog



'''
Parse ris files into dict format, and export DocMeta dict to ris file.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import re
import logging
from pprint import pprint
from RISparser import readris
from RISparser.config import LIST_TYPE_TAGS, TAG_KEY_MAPPING
try:
    from . import sqlitedb
except:
    import sqlitedb

LOGGER=logging.getLogger(__name__)

ALT_KEYS={
        'keywords': 'keywords_l',
        'tag': 'tags_l', # probably no such thing
        'url': 'urls_l'
        }

INV_ALT_KEYS=dict([(vv,kk) for kk,vv in ALT_KEYS.items()])

TYPE_DICT={'Report': 'RPRT',\
           'JournalArticle': 'JOUR',\
           'article': 'JOUR',\
           'Book': 'BOOK',\
           'BookSection': 'CHAP',\
           'ConferenceProceedings': 'CONF',\
           'Generic': 'GEN',\
           'Bill': 'BILL',\
           'Case': 'CASE',\
           'ComputerProgram': 'COMP',\
           'EncyclopediaArticle': 'ENCYC',\
           'Film': 'VIDEO',\
           'Hearing': 'HEAR',\
           'MagazineArticle': 'MGZN',\
           'NewspaperArticle': 'NEWS',\
           'Patent': 'PAT',\
           'Statute': 'STAT',\
           'Thesis': 'THES',\
           'WebPage': 'ELEC',\
           'WorkingPaper': 'MANSCPT'}


KEYWORD_DICT={'title': 'TI',\
              'issue': 'IS',\
              'publication': 'JO',\
              'volume': 'VL',\
              'doi': 'DO',\
              'abstract': 'AB',\
              'edition': 'ET',\
              'ISBN': 'SN',\
              'isbn': 'SN',\
              'ISSN': 'SN',\
              'issn': 'SN',\
              'publisher': 'PB',\
              #'keywords': 'KW',\
              #'path': 'L1',\
              'notes': 'N1',\
              'editor': 'ED'}



def splitNames(entry):

    firstnames=[]
    lastnames=[]
    for nii in entry['authors']:
        lii,fii=nii.split(',',1)
        firstnames.append(fii.strip())
        lastnames.append(lii.strip())

    entry['firstNames_l']=firstnames
    entry['lastName_l']=lastnames

    return entry


def readRISFile(filename):
    """Read and parse RIS file

    Args:
        filename (str): abspath to ris file.

    Returns: results (list): list of dicts of meta data.
    """

    filename=os.path.abspath(filename)
    if not os.path.exists(filename):
        return None

    results=[]
    with open(filename,'r') as fin:

        entries=readris(fin)
        LOGGER.info('Read in RIS file: %s' %filename)

        for eii in entries:
            #eii=altKeys(eii,ALT_KEYS)

            # type
            if eii['type_of_reference']=='JOUR':
                eii['type']='article'
            if eii['type_of_reference']=='BOOK':
                eii['type']='book'

            # authors
            if 'authors' not in eii:
                authors=[]
                for kii in ['first_authors', 'secondary_authors',
                        'tertiary_authors', 'subsidiary_authors']:
                    aii=eii.get(kii)
                    if aii:
                        authors.extend(aii)
                eii['authors']=authors
            eii=splitNames(eii)

            LOGGER.debug('authors = %s' %eii['authors'])

            # title
            title=eii.get('title')
            if not title:
                for tii in ['primary_title', 'secondary_title', 'tertiary_title',
                        'translated_title']:
                    if tii in eii:
                        eii['title']=eii[tii]
                        break

            LOGGER.debug('title = %s' %eii['title'])

            # citationkey
            citationkey=eii.get('id')
            if citationkey:
                eii['citationkey']=citationkey

            LOGGER.debug('citationkey = %s' %eii['citationkey'])

            # date
            if 'date' in eii:
                year,month,day=eii['date'].split('/')
                eii['year']=year
                eii['month']=month
                eii['day']=day

            LOGGER.debug('year = %s. month = %s. day = %s' %(year,month,day))

            # pages
            sp=eii.get('start_page', 'n/a')
            ep=eii.get('end_page', 'n/a')
            eii['pages']='%s-%s' %(sp,ep)

            # publication
            if 'journal_name' in eii and 'publication' not in eii:
                eii['publication']=eii['journal_name']

            LOGGER.debug('publication = %s' %eii['publication'])

            # keywords
            kw=eii.get('keywords')
            if kw:
                eii['keywords_l']=kw

            # files
            f1=eii.get('file_attachments1')
            f2=eii.get('file_attachments2')
            files=[fii for fii in [f1,f2] if fii]
            eii['files_l']=files

            # notes
            notes=eii.get('notes')
            if notes:
                eii['notes']='\n\n'.join(notes)

            # urls
            urls=eii.get('url')
            if urls:
                eii['urls_l']=[urls,]

            docii=sqlitedb.DocMeta()
            docii.update(eii)
            results.append(docii)

    return results


def metaDictToRIS(jobid, metadict, path_prefix):
    """A wrapper, export meta data dict to RIS format

    Args:
        jobid (int): job id.
        metadict (dict): meta data dict.
        path_prefix (str): path prefix to prepend to attachment files.

    Returns:
        rec (int): 0 if success, 1 otherwise.
        jobid (int): input jobid.
        docid (int): 'id' value in <metadict>
    """

    try:
        text=parseMeta(metadict,path_prefix)
        return 0,jobid,text,metadict['id']
    except Exception:
        LOGGER.exception('Failed to write to RIS format. Jobid = %s. Doc id = %s'\
                %(str(jobid), metadict['id']))
        return 1,jobid,'',metadict['id']


def parseMeta(metadict, path_prefix):
    """Export meta data dict to RIS format

    Args:
        metadict (dict): meta data dict.
        path_prefix (str): path prefix to prepend to attachment files.

    Returns: string (str): RIS text.
    """

    def getField(doc,field,default=''):
        return doc[field] or default

    page_re=re.compile('(.*?)-+(.*)', re.UNICODE)
    def _subDash(match):
        return '%s--%s' %(match.group(1),match.group(2))

    #--------------------Get type--------------------
    doctype=getField(metadict,'type','article')
    doctype=TYPE_DICT[doctype]
    entries=['TY - %s' %doctype,]

    LOGGER.debug('doctype (TY) = %s' %doctype)

    #-------------------Get authors-------------------
    authors=sqlitedb.zipAuthors(metadict['firstNames_l'], metadict['lastName_l'])
    for aii in authors:
        entries.append('AU - %s' %aii)

        LOGGER.debug('authors (AU) = %s' %aii)
    #authors=latexencode.utf8tolatex(authors)

    #----------------------Get id----------------------
    citationkey=metadict['citationkey']
    if citationkey:
        entries.append('ID - %s' %citationkey)

        LOGGER.debug('citationkey (ID) = %s' %citationkey)

    #---------------------Get time---------------------
    year=getField(metadict,'year','')
    month=getField(metadict,'month','')
    day=getField(metadict,'day','')
    time=[]
    for ii in [year,month,day]:
        try:
            ii=str(int(ii))
        except:
            # vv is nan
            ii=''
        time.append(ii)
    if year!='':
        entries.append('PY - %s' %time[0])
    time='%s/%s/%s/' %(time[0],time[1],time[2])
    entries.append('DA - %s' %time)
    entries.append('Y1 - %s' %time)

    LOGGER.debug('time (PY, DA, Y1) = %s' %time)

    #--------------------Get pages--------------------
    pages=getField(metadict,'pages','')
    if pages!='':
        pmatch=page_re.match(pages)
        if pmatch is None:
            entries.append('SP - %s' %str(pages))
            LOGGER.debug('pages (SP) = %s' %(str(pages)))
        else:
            entries.append('SP - %s' %str(pmatch.group(1)))
            entries.append('EP - %s' %str(pmatch.group(2)))

            LOGGER.debug('pages (SP) = %s' %(str(pmatch.group(1))))
            LOGGER.debug('pages (EP) = %s' %(str(pmatch.group(2))))


    #-----------------Get city/country-----------------
    loc=''
    city=getField(metadict,'city','')
    country=getField(metadict,'country','')
    if city!='':
        loc=u'%s, %s' %(loc,city)
    if country!='':
        loc=u'%s, %s' %(loc,country)
    if len(loc)>0:
        entries.append('CY - %s' %loc)

    LOGGER.debug('city (CY) = %s' %loc)

    #------------------Get file path------------------
    files=metadict['files_l']
    if files:
        # can only store 2 files?
        for fii in files[:2]:
            entries.append('L1 - %s' %os.path.join(path_prefix,fii))

            LOGGER.debug('file (L1) = %s' %fii)

    #--------------Populate other fields--------------
    for kk,vv in metadict.items():
        if vv is None:
            continue
        if kk in ['type','firstnames','lastname','docid','year','month',\
                'day','pages','city','country', 'path']:
            continue

        if kk =='keywords_l' and len(vv)>0:
            for kii in vv:
                entries.append('KW - %s' %kii)

                LOGGER.debug('keywords (KW) = %s' %kii)

	#-----------Specifiy issn and isbn-----------------
        if kk.lower()=='issn':
            vv='issn %s' %vv
        if kk.lower()=='isbn':
            vv='isbn %s' %vv

        #--------------------All others--------------------
        kk=KEYWORD_DICT.get(kk,None)
        if kk is None:
            continue
            entries.append('%s - %s' %(kk,vv))

            LOGGER.debug('other key (%s) = %s' %(kk,vv))

    entries.append('ER -\n')
    string='\n'.join(entries)

    LOGGER.info('Done writing to RIS format')

    return string




if __name__=='__main__':

    filename='../testris4.ris'

    '''
    print('list tags',LIST_TYPE_TAGS)

    with open(filename,'r') as fin:
        entries=readris(fin)
        for eii in entries:
            pprint(eii)
    '''
    results=readRISFile(filename)
    km=TAG_KEY_MAPPING

    for ii in results:
        print('#################################\n')
        pprint(ii)

        #rii=parseMeta(ii)
        #pprint(rii)
        rii=metaDictToRIS(0,ii)
        print(rii)




'''
MeiTing Trunk

An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.


Slots involving folder changes, including creation, re-parenting, renaming,
trashing, and restoring from trash.
'''

from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot
from PyQt5 import QtWidgets
from PyQt5.QtGui import QCursor, QBrush, QColor, QIcon
from lib import sqlitedb
from lib.tools import iterTreeWidgetItems



class MainFrameLibTreeSlots:

    #######################################################################
    #                            Libtree slots                            #
    #######################################################################


    def clickSelFolder(self, item, column):
        """Change button states and load doc table on selecting a folder

        Args:
            item (QTreeWidgetItem): selected item in folder tree.
            column (int): column index in folder tree.
        """

        folder=item.data(0,0)
        folderid=item.data(1,0)

        # NOTE that this hiding is necessary, although it is called via
        # filterTypeCombboxChange() -> clearFilterButtonClicked().
        # Because in that case, the loadDocTable() will be called twice,
        # if the clear_filter_frame was visible before this clickSelFolder()
        # call. If I remove loadDocTable() and leave it in the
        # clearFilterButtonClicked(), it would be less transparent.
        if self.clear_filter_frame.isVisible():
            self.clear_filter_frame.setVisible(False)

        self.logger.info('Selected folder = %s. folderid = %s' \
                %(folder, folderid))

        if item==self.all_folder:
            self.loadDocTable(folder=None,sortidx=4,sel_row=0)
        else:
            self.loadDocTable((folder,folderid),sortidx=4,sel_row=0)

        if item==self.all_folder:
            self.add_button.setDisabled(True)
            self.create_subfolder_action.setDisabled(True)
            self.create_folder_action.setEnabled(True)
            self.add_folder_button.setEnabled(True)
            self.duplicate_check_button.setEnabled(True)
            self.search_button.setEnabled(True)
        elif item==self.needsreview_folder:
            self.add_button.setDisabled(True)
            self.add_folder_button.setDisabled(True)
            self.search_button.setEnabled(True)
        elif item==self.trash_folder:
            self.add_button.setDisabled(True)
            self.add_folder_button.setDisabled(True)
            self.search_button.setEnabled(True)
        else:
            if folderid in self._trashed_folder_ids:
                self.add_button.setDisabled(True)
                self.add_folder_button.setDisabled(True)
                self.create_subfolder_action.setDisabled(True)
                self.search_button.setEnabled(True)
            else:
                self.add_button.setEnabled(True)
                self.add_folder_button.setEnabled(True)
                self.create_subfolder_action.setEnabled(True)
                self.duplicate_check_button.setEnabled(True)
                self.search_button.setEnabled(True)

        # Refresh filter list
        self.filterTypeCombboxChange()

        return


    @pyqtSlot()
    def selFolder(self):
        '''Select folder by changing current

        This is the slot to libtree.selectionModel().selectionChanged signal.
        '''

        item=self._current_folder_item
        if item:
            self.clickSelFolder(item,0)

        return


    def libTreeMenu(self,pos):
        '''Right click menu in folder tree'''

        item=self._current_folder_item
        folderid=item.data(1,0)

        if item:
            menu=QtWidgets.QMenu()
            if item==self.trash_folder or folderid in self._trashed_folder_ids:
                menu_type='trash'
                restore_action=menu.addAction('Restore Folder(s)')
                clear_action=menu.addAction('Delete From Trash')
            else:
                menu_type='default'
                add_action=menu.addAction('&Create Folder')
                add_action.setIcon(self.style().standardIcon(
                    QtWidgets.QStyle.SP_DirOpenIcon))
                add_action.setShortcut('C')

                addsub_action=menu.addAction('Create &Sub Folder')
                addsub_action.setIcon(self.style().standardIcon(
                    QtWidgets.QStyle.SP_DirClosedIcon))
                addsub_action.setShortcut('S')

                del_action=menu.addAction('&Delete Folder')
                del_action.setIcon(self.style().standardIcon(
                    QtWidgets.QStyle.SP_TrashIcon))
                del_action.setShortcut('D')

                rename_action=menu.addAction('&Rename Folder')
                rename_action.setIcon(QIcon.fromTheme('edit-select-all'))
                rename_action.setShortcut('R')

                if item==self.needsreview_folder:
                    add_action.setEnabled(False)
                    addsub_action.setEnabled(False)
                    del_action.setEnabled(False)
                    rename_action.setEnabled(False)
                elif item==self.all_folder:
                    addsub_action.setEnabled(False)
                    del_action.setEnabled(False)
                    rename_action.setEnabled(False)

            action=menu.exec_(QCursor.pos())

            if action:

                self.logger.info('action.text() = %s' %action.text())

                if menu_type=='trash':
                    if action==restore_action:
                        self.restoreFolderFromTrash()
                    elif action==clear_action:
                        self.delFolderFromTrash(item)
                else:
                    if action==add_action:
                        self.addFolderButtonClicked(add_action)
                    elif action==addsub_action:
                        self.addFolderButtonClicked(addsub_action)
                    elif action==del_action:
                        self.trashFolder(item,None,True)
                    elif action==rename_action:
                        self.renameFolder()

        return


    @pyqtSlot(str,str)
    def changeFolderParent(self, move_folder_id, new_parent_id):
        """Change the parent of a folder

        Args:
            move_folder_id (str): id of changed folder.
            new_parent_id (str): id of new parent folder.

        This is the slot to libree.folder_move_signal signal, which is emitted
        on re-parenting by drag/drop. See lib/widgets/folder_tree.py.

        It is also called when trashing a folder, see trashFolder().
        """

        folder_name=self.folder_dict[move_folder_id][0]

        #------------------Restoring docs------------------
        trashed_folder_ids=self._trashed_folder_ids
        if move_folder_id in trashed_folder_ids and new_parent_id \
                not in trashed_folder_ids:
            self.logger.info('Moving folder (id = %s) out from Trash. Restoring docs within.'\
                    %move_folder_id)
            for docid in self.folder_data[move_folder_id]:
                self.meta_dict[docid]['deletionPending']=='false'
                self.changed_doc_ids.append(docid)

                self.logger.debug("Restoring doc %d. meta_dict[docid]['deletionPending'] = %s"\
                        %(docid, self.meta_dict[docid]['deletionPending']))

        self.folder_dict[move_folder_id]=(folder_name, new_parent_id)

        self.logger.debug('folder_dict[move_folder_id] = %s'\
                %str(self.folder_dict[move_folder_id]))

        self.changed_folder_ids.append(move_folder_id)

        return


    @pyqtSlot(QtWidgets.QTreeWidgetItem,QtWidgets.QTreeWidgetItem,bool)
    def trashFolder(self, item, newparent=None, ask=True):
        """Send a folder into Trash

        Args:
            item (QTreeWidgetItem): item of the folder to trash.

        Kwargs:
            newparent (QTreeWidgetItem or None): if QTreeWidgetItem, the new
                parent item onto which to add <item>. This happens when
                dragging a folder onto a folder already in Trash, or onto Trash
                itself.  Then this function acts as the slot to
                libtree.folder_del_signal emitted in libtree.dropEvent(), see
                lib/widgets/folder_tree.py

                If None, send the folder to Trash, this happens when trashing
                a folder from the right click menu, see libTreeMenu().
            ask (bool): whether to prompt for confirmation. Default to True.
                It gets overwritten to False if it's moving folders within
                the Trash.
        """

        folderid=item.data(1,0)
        do_post_trash=True

        #--------------Trashing by drag/drop--------------
        if newparent is not None:

            if folderid in self._trashed_folder_ids+['-3']:
                # move within Trash
                self.logger.debug('newparent id %s, folderid %s both in _trashed_folder_ids. Skip postTrashFolder()'\
                        %(newparent.data(1,0), folderid))
                ask=False
                do_post_trash=False

        if ask:
            choice=QtWidgets.QMessageBox.question(self, 'Confirm deletion',
                'Deleting a folder will delete all sub-folders and documents inside.\n\nConfirm?',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        if not ask or (ask and choice==QtWidgets.QMessageBox.Yes):

            root=self.libtree.invisibleRootItem()
            (item.parent() or root).removeChild(item)

            if newparent is None:
                self.logger.info('Send folder %s to Trash' %folderid)
                self.trash_folder.addChild(item)
                self.changeFolderParent(folderid,'-3')
            else:
                self.logger.info('Put folder %s to trashed folder %s'\
                        %(folderid, newparent.data(1,0)))
                newparent.addChild(item)
                self.changeFolderParent(folderid,newparent.data(1,0))

            if do_post_trash:
                self.postTrashFolder(item)

        return


    def postTrashFolder(self, item):
        """Update orphan doc records after folder trashing

        Args:
            item (QTreeWidgetItem): item of folder been trashed.
        """

        folderid=item.data(1,0)

        # get subfolders and docs in trashed folder and subfolders
        delfolderids,deldocids=sqlitedb.walkFolderTree(self.folder_dict,
                self.folder_data,folderid)
        # find orphan docs within
        orphan_docs=sqlitedb.findOrphanDocs(self.folder_data,deldocids,
                self._trashed_folder_ids)

        self.logger.debug('Ids of folders to trash = %s' %delfolderids)
        self.logger.debug('Ids of docs within = %s' %deldocids)
        self.logger.info('Orphan docs = %s' %orphan_docs)

        for idii in orphan_docs:
            self.meta_dict[idii]['deletionPending']='true'
            self.logger.debug('Set deletionPending to orphan doc %s %s' \
                    %(idii, self.meta_dict[idii]['deletionPending']))

        self.changed_doc_ids.extend(orphan_docs)

        return


    def restoreFolderFromTrash(self):

        msg=QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        msg.setWindowTitle('Restore')
        msg.setText('Not implemented yet.')
        msg.setInformativeText('You can drag/drop trashed folders/documents to restore them.')
        msg.exec_()

        return


    def delFolderFromTrash(self, item):
        """Delete a trashed folder

        Args:
            item (QTreeWidgetItem): item of folder to be deleted
        """

        choice=QtWidgets.QMessageBox.question(self, 'Confirm deletion',
            'Deleting folder(s) and document within permanently?',
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if choice==QtWidgets.QMessageBox.No:
            return

        foldername=item.data(0,0)
        folderid=item.data(1,0)
        item_parent=item.parent()

        self.logger.info('current folder name = %s. folder id = %s.' \
                %(foldername, folderid))

        # Not sure, may need to stop the timer
        self.auto_save_timer.stop()
        self.logger.info('Stoped auto save timer.')

        #-------------------Empty trash-------------------
        if folderid=='-3':
            changed_folders=self._trashed_folder_ids

            #-----------------Get orphan docs in Trash-----------------
            for docii in self.folder_data['-3']:

                self.logger.warning('Deleting orphan doc %s from meta_dict' %docii)
                del self.meta_dict[docii]
                self.logger.warning('Deleting orphan doc %s from folder_data[-3]' %docii)
                self.folder_data['-3'].remove(docii)
                self.changed_doc_ids.append(docii)

        #-------------------Empty folders in trash-------------------
        else:
            changed_folders, _=sqlitedb.walkFolderTree(self.folder_dict,
                    self.folder_data, folderid)

        self.logger.info('Ids of folders to clear from trash = %s' %changed_folders)

        #-------------Destroy docs in folders-------------
        for fii in changed_folders:
            self.logger.info('Destroying docs in folder %s' %fii)
            self.destroyDoc(self.folder_data[fii], fii, False, False)

        #-----Del folders after all docs are destroyed-----
        for fii in changed_folders:
            self.logger.warning('Deleting folder %s from folder_dict, folder_data' %fii)
            del self.folder_dict[fii]
            del self.folder_data[fii]

            itemii=self.libtree.findItems(fii, Qt.MatchExactly | Qt.MatchRecursive,
                    column=1)
            if len(itemii)>0:
                itemii=itemii[0]
                itemii.parent().removeChild(itemii)

        self.changed_folder_ids.extend(changed_folders)

        self.libtree.setCurrentItem(item_parent)

        self.auto_save_timer.start()
        self.logger.info('Restarted auto save timer.')

        return


    def renameFolder(self):

        item=self._current_folder_item
        if item:
            if item in self.sys_folders:
                return
            item.setFlags(item.flags() | Qt.ItemIsEditable)
            self.libtree.scrollToItem(item)
            self.libtree.editItem(item)

        return


    def removeFolderHighlights(self):

        ori_color=QBrush(QColor(255,255,255))
        root=self.libtree.invisibleRootItem()
        for item in iterTreeWidgetItems(self.libtree, root):
            item.setBackground(0, ori_color)

        return


